###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# toolbox/                                                                    #
#         cuckoo.py                                                           #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


from future.builtins import super  # pylint: disable=wrong-import-order

import datetime
import re
import os
import locale
import logging
import json
import subprocess
import random
import requests
import urllib3.util.retry

from threading import RLock, Event
from time import sleep
from twisted.internet import protocol, reactor, process

from peekaboo.exceptions import CuckooSubmitFailedException


logger = logging.getLogger(__name__)


class CuckooJob(object):
    """ Remember sample and submission time of a Cuckoo job. """
    def __init__(self, sample):
        self.__sample = sample
        self.__submission_time = datetime.datetime.utcnow()

    def is_older_than(self, seconds):
        """ Returns True if the difference between submission time and now,
        i.e. the age of the job, is larger than given number of seconds. """
        max_age = datetime.timedelta(seconds=seconds)
        return datetime.datetime.utcnow() - self.__submission_time > max_age

    @property
    def sample(self):
        """ Returns the sample the job is analyzing. """
        return self.__sample


class Cuckoo(object):
    """ Parent class, defines interface to Cuckoo. """
    def __init__(self, job_queue):
        self.job_queue = job_queue
        self.shutdown_requested = Event()
        self.shutdown_requested.clear()
        self.running_jobs = {}
        # reentrant because we're doing nested calls within critical sections
        self.running_jobs_lock = RLock()

    def register_running_job(self, job_id, sample):
        """ Register a job as running. Detect if another sample has already
        been registered with the same job ID which obviously must never happen
        because it corrupts our internal housekeeping. Guarded by a lock
        because multiple worker threads will call this routine and check for
        collision and update of job log might otherwise race each other.

        @param job_id: ID of the job to register as running.
        @type job_id: int
        @param sample: Sample object to associate with this job ID
        @type sample: Sample

        @returns: None
        @raises: CuckooSubmitFailedException on job id collision """
        with self.running_jobs_lock:
            if (job_id in self.running_jobs and
                    self.running_jobs[job_id] is not sample):
                raise CuckooSubmitFailedException(
                    'A job with ID %d is already registered as running '
                    'for sample %s' % (job_id, self.running_jobs[job_id]))

            self.running_jobs[job_id] = CuckooJob(sample)

    def deregister_running_job(self, job_id):
        """ Deregister a running job by job id.

        @returns: Sample object of the job or None if job not found. """
        with self.running_jobs_lock:
            job = self.running_jobs.pop(job_id, None)
            if job is not None:
                return job.sample

        return None

    def deregister_running_job_if_too_old(self, job_id, max_age):
        """ Check if a job has gotten too old and remove it from the list of
        running jobs if so.

        @returns: Sample object of the job or None if job not found. """
        with self.running_jobs_lock:
            if self.running_jobs[job_id].is_older_than(max_age):
                return self.deregister_running_job(job_id)

        return None

    def resubmit_with_report(self, job_id):
        """ Resubmit a sample to the job queue after the report became
        available. Retrieves the report from Cuckoo.

        @param job_id: ID of job which has finished.
        @type job_id: int

        @returns: None """
        logger.debug("Analysis done for task #%d" % job_id)

        sample = self.deregister_running_job(job_id)
        if sample is None:
            logger.debug('No sample found for job ID %d', job_id)
            return None

        logger.debug('Requesting Cuckoo report for sample %s', sample)
        report = self.get_report(job_id)
        if report is None:
            # mark analysis as failed if we could not get the report e.g.
            # because it was corrupted or the API connection failed.
            sample.mark_cuckoo_failure()
        else:
            reportobj = CuckooReport(report)
            sample.register_cuckoo_report(reportobj)

        self.job_queue.submit(sample, self.__class__)
        return None

    def resubmit_as_failed_if_too_old(self, job_id, max_age):
        """ Resubmit a sample to the job queue with a failure report if the
        Cuckoo job has been running for too long.

        @param job_id: ID of job to check.
        @type job_id: int
        @param max_age: maximum job age in seconds
        @type max_age: int
        """
        sample = self.deregister_running_job_if_too_old(job_id, max_age)
        if sample is not None:
            logger.warning("Dropped job %d because it has been running for "
                           "too long", job_id)
            sample.mark_cuckoo_failure()
            self.job_queue.submit(sample, self.__class__)

    def shut_down(self):
        """ Request the module to shut down. """
        self.shutdown_requested.set()

    def reap_children(self):
        pass

    def get_report(self, job_id):
        """ Extract the report of a finished analysis from Cuckoo. To be
        overridden by derived classes for actual implementation. """
        raise NotImplementedError

class CuckooEmbed(Cuckoo):
    """ Runs and interfaces with Cuckoo in IPC. """
    def __init__(self, job_queue, cuckoo_exec, cuckoo_submit,
                 cuckoo_storage, interpreter=None):
        super().__init__(job_queue)
        self.interpreter = interpreter
        self.cuckoo_exec = cuckoo_exec
        self.cuckoo_submit = cuckoo_submit
        self.cuckoo_storage = cuckoo_storage
        self.exit_code = 0

        # process output to get job ID
        patterns = (
            # Example: Success: File "/var/lib/peekaboo/.bashrc" added as task with ID #4
            "Success.*: File .* added as task with ID #([0-9]*)",
            "added as task with ID ([0-9]*)",
        )
        self.job_id_patterns = [re.compile(pattern) for pattern in patterns]

    def submit(self, sample):
        """
        Submit a file or directory to Cuckoo for behavioural analysis.

        @param sample: Sample object to analyse.
        @return: The job ID used by Cuckoo to identify this analysis task.
        """
        try:
            # cuckoo_submit is a list, make a copy as to not modify the
            # original value
            proc = self.cuckoo_submit.split(' ') + [sample.submit_path]

            # universal_newlines opens channels to child in text mode and
            # returns strings instead of bytes in return which we do to avoid
            # the need to handle decoding ourselves
            p = subprocess.Popen(proc,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 universal_newlines=True)
            p.wait()
        except Exception as error:
            raise CuckooSubmitFailedException(error)

        if not p.returncode == 0:
            raise CuckooSubmitFailedException(
                'cuckoo submit returned a non-zero return code.')

        out, err = p.communicate()
        logger.debug("cuckoo submit STDOUT: %s", out)
        logger.debug("cuckoo submit STDERR: %s", err)

        match = None
        pattern_no = 0
        for pattern in self.job_id_patterns:
            match = re.search(pattern, out)
            if match is not None:
                logger.debug('Pattern %d matched.', pattern_no)
                break

            pattern_no += 1

        if match is not None:
            job_id = int(match.group(1))
            self.register_running_job(job_id, sample)
            return job_id

        raise CuckooSubmitFailedException(
            'Unable to extract job ID from given string %s' % out)

    def get_report(self, job_id):
        path = os.path.join(self.cuckoo_storage,
                'analyses/%d/reports/report.json' % job_id)

        if not os.path.isfile(path):
            return None

        logger.debug('Accessing Cuckoo report for task %d at %s ' %
                (job_id, path))

        report = None
        with open(path) as data:
            try:
                report = json.load(data)
            except ValueError as e:
                logger.warning("Error loading JSON report for cuckoo "
                               "job id %d", job_id)

        return report

    def do(self):
        """ Run Cuckoo sandbox, parse log output, and report back of Peekaboo. """
        command = self.cuckoo_exec.split(' ')

        # allow for injecting a custom interpreter which we use to run cuckoo
        # with python -u for unbuffered standard output
        if self.interpreter:
            command = self.interpreter.split(' ') + command

        reactor.spawnProcess(CuckooServer(self), command[0], command)

        # do not install twisted's signal handlers because it will screw with
        # our logic (install a handler for SIGTERM and SIGCHLD but not for
        # SIGINT). Instead do what their SIGCHLD handler would do and call the
        # global process reaper.
        reactor.run(installSignalHandlers = False)
        process.reapAllProcesses()
        return self.exit_code

    def shut_down(self, exit_code = 0):
        """ Signal handler callback but in this instance also used as callback
        for protocol to ask us to shut down if anything adverse happens to the
        child """
        # the reactor doesn't like it to be stopped more than once and catching
        # the resulting ReactorNotRunning exception is foiled by the fact that
        # sigTerm defers the call through a queue into another thread which
        # insists on logging it
        if not self.shutdown_requested.is_set():
            reactor.sigTerm(0)

        self.shutdown_requested.set()
        self.exit_code = exit_code

    def reap_children(self):
        """ Since we only have one child, SIGCHLD will cause us to shut down
        and we reap all child processes on shutdown. This method is therefore
        (currently) intentionally a no-op. """
        pass


class WhitelistRetry(urllib3.util.retry.Retry):
    """ A Retry class which has a status code whitelist, allowing to retry all
    requests not whitelisted in a hard-core, catch-all manner. """
    def __init__(self, status_whitelist=None, **kwargs):
        super().__init__(**kwargs)
        self.status_whitelist = status_whitelist or set()

    def is_retry(self, method, status_code, has_retry_after=False):
        """ Override Retry's is_retry to introduce our status whitelist logic.
        """
        # we retry all methods so no check if method is retryable here

        if self.status_whitelist and status_code not in self.status_whitelist:
            return True

        return super().is_retry(method, status_code, has_retry_after)


class CuckooApi(Cuckoo):
    """ Interfaces with a Cuckoo installation via its REST API. """
    def __init__(self, job_queue, url="http://localhost:8090", api_token="",
                 poll_interval=5, submit_original_filename=True,
                 max_job_age=900, retries=5, backoff=0.5):
        super().__init__(job_queue)
        self.url = url
        self.api_token = api_token
        self.poll_interval = poll_interval
        self.submit_original_filename = submit_original_filename
        self.max_job_age = max_job_age

        # urrlib3 backoff formula:
        # <backoff factor> * (2 ^ (<retry count so far> - 1))
        # with second try intentionally having no sleep,
        # e.g. with retry count==5 and backoff factor==0.5:
        # try 1: fail, sleep(0.5*2^(1-1)==0.5*2^0==0.5*1==0.5->intentionally
        #   overridden to 0)
        # try 2: fail, sleep(0.5*2^(2-1)==0.5*2^1==1)
        # try 3: fail, sleep(0.5*2^(3-1)==0.5*2^2==2)
        # try 4: fail, sleep(0.5*2^(4-1)==0.5*2^3==4)
        # try 5: fail, abort, sleep would've been 8 before try 6
        #
        # Also, use method_whitelist=False to enable POST and other methods for
        # retry which aren't by default because they're not considered
        # idempotent. We assume that with the REST API a request either
        # succeeds or fails without residual effects, making them atomic and
        # idempotent.
        #
        # And finally we retry everything but a 200 response, which admittedly
        # is a bit hard-core but serves our purposes for now.
        retry_config = WhitelistRetry(total=retries,
                                      backoff_factor=backoff,
                                      method_whitelist=False,
                                      status_whitelist=set([200]))
        retry_adapter = requests.adapters.HTTPAdapter(max_retries=retry_config)
        self.session = requests.session()
        self.session.mount('http://', retry_adapter)
        self.session.mount('https://', retry_adapter)

    def __get(self, path):
        request_url = "%s/%s" % (self.url, path)
        logger.debug("Getting %s", request_url)
        headers = {"Authorization": "Bearer %s" % self.api_token}

        try:
            response = self.session.get(request_url, headers=headers)
        # all requests exceptions are derived from RequestsException, including
        # RetryError, TooManyRedirects and Timeout
        except requests.exceptions.RequestException as error:
            logger.error('Request to REST API failed: %s', error)
            return None

        # no check for status code here since we retry all but 200
        # responses and raise an exception if retries fail
        try:
            json_resp = response.json()
        except ValueError as error:
            logger.error(
                'Invalid JSON in response when getting %s: %s',
                request_url, error)
            return None

        return json_resp

    def submit(self, sample):
        path = sample.submit_path
        filename = os.path.basename(path)
        # override with the original file name if available
        if self.submit_original_filename:
            if sample.name_declared:
                filename = sample.name_declared
            elif sample.filename:
                filename = sample.filename

        files = {"file": (filename, open(path, 'rb'))}
        logger.debug("Creating Cuckoo task with content from %s and "
                     "filename %s", path, filename)
        headers = {"Authorization": "Bearer %s" % self.api_token}

        try:
            response = self.session.post(
                "%s/tasks/create/file" % self.url, headers=headers, files=files)
        except requests.exceptions.RequestException as error:
            raise CuckooSubmitFailedException(
                'Error creating Cuckoo task: %s' % error)

        try:
            json_resp = response.json()
        except ValueError as error:
            raise CuckooSubmitFailedException(
                'Invalid JSON in response when creating Cuckoo task: %s'
                % error)

        if "task_id" in json_resp:
            task_id = json_resp["task_id"]
            if task_id > 0:
                self.register_running_job(task_id, sample)
                return task_id

        raise CuckooSubmitFailedException(
            'Unable to extract job ID from response %s' % json_resp)

    def get_report(self, job_id):
        logger.debug("Report from Cuckoo API requested, job_id = %d" % job_id)
        return self.__get("tasks/report/%d" % job_id)

    def do(self):
        """ Do the polling for finished jobs. """
        # do a simple initial test of connectivity. With this we require the
        # API to be reachable at startup (with the usual retries to account for
        # a bit of a race condition in parallel startup) but later on hope
        # that all errors are transient and retry endlessly
        status = self.__get("cuckoo/status")
        if status is None:
            logger.critical("Connection to Cuckoo REST API failed")
            return 1
        if "tasks" not in status or "reported" not in status["tasks"]:
            logger.critical("Invalid status JSON structure from Cuckoo REST "
                            "API: %s", status)
            return 1

        reported = status["tasks"]["reported"]
        logger.info("Connection to Cuckoo seems to work, "
                    "%i reported tasks seen", reported)

        while not self.shutdown_requested.wait(self.poll_interval):
            # no lock, atomic, copy() because keys() returns an iterable view
            # instead of a fresh new list in python3
            running_jobs = self.running_jobs.copy().keys()

            # somewhat inefficient for the potential number of requests per
            # poll round but more efficient in that it does not download an
            # ever growing list of jobs using tasks/list. Might also take
            # quite a while to get to all jobs if retries happen on each
            # request.
            # A call to get data about multiple tasks in one go would be nice
            # here. tasks/list could be used with the minimum job number as
            # offset and spread between highest and lowest job id as limit *if*
            # its output was sorted by job ID. Apparently limit and offset are
            # only meant to iterate over the job list in blocks but not to
            # return data about a specific range of job IDs from that list.
            for job_id in running_jobs:
                job = self.__get("tasks/view/%i" % job_id)
                if job is None:
                    # ignore and retry on next polling run
                    continue

                # but fail hard if we get invalid stuff
                if "task" not in job or "status" not in job["task"]:
                    logger.error("Invalid JSON structure from Cuckoo REST "
                                 "API: %s", job)
                    return 1

                if job["task"]["status"] == "reported":
                    self.resubmit_with_report(job_id)
                    continue

                # drop jobs which have been running for too long. This is
                # mainly to prevent accumulation of jobs in our job list which
                # will never finish. We still want to wait for jobs to finish
                # even though our client might not be interested any more so
                # that we have the result cached for the next time we get the
                # same sample.
                self.resubmit_as_failed_if_too_old(job_id, self.max_job_age)

        logger.debug("Shutting down.")
        return 0

class CuckooServer(protocol.ProcessProtocol):
    """ Class that is used by twisted.internet.reactor to process Cuckoo output
    and process its behavior. Usage::

        srv = CuckooServer()
        reactor.spawnProcess(srv, 'python2', ['python2', '/path/to/cukoo.py'])
        reactor.run() """
    def __init__(self, cuckoo):
        self.cuckoo = cuckoo
        self.encoding = locale.getpreferredencoding()

    def connectionMade(self):
        logger.info('Connected. Cuckoo PID: %s' % self.transport.pid)
        return None

    def outReceived(self, data):
        """ on receiving output on STDOUT from Cuckoo """
        # explicit decoding: The program is sending us stuff and because it's
        # just stdout/stderr we have no defined protocol, no structure and no
        # guaranteed encoding. Normally we'd tell popen to open in text mode
        # which would automatically apply the system encoding. With Twisted
        # there doesn't seem to be that option. But since it's our child, we
        # can (hopefully) assume that it uses our locale settings. So we use
        # the default encoding as returned by our interpreter.
        logger.debug('STDOUT %s', data.decode(self.encoding))

    def errReceived(self, data):
        """ on receiving output on STDERR from Cuckoo """
        content = data.decode(self.encoding)
        logger.debug('STDERR %s', content.replace('\n', ''))

        #
        # FILE SUBMITTED
        # printed out but has no further effect
        #
        # 2016-04-12 09:14:06,984 [lib.cuckoo.core.scheduler] INFO: Starting
        # analysis of FILE "cuckoo.png" (task #201, options "")
        # INFO: Starting analysis of FILE ".bashrc" (task #4, options "")
        match = re.match(r'.*INFO: Starting analysis of FILE "(.*)" '
                         r'\(task #([0-9]*), options .*', content)

        if match:
            logger.info("File submitted: task #%s, filename %s",
                        match.group(2), match.group(1))

        #
        # ANALYSIS DONE
        #
        # 2016-04-12 09:25:27,824 [lib.cuckoo.core.scheduler] INFO: Task #202:
        # reports generation completed ...
        m = re.match(".*INFO: Task #([0-9]*): reports generation completed.*",
                     content)
        if m:
            job_id = int(m.group(1))
            self.cuckoo.resubmit_with_report(job_id)

    def inConnectionLost(self):
        logger.debug("Cuckoo closed STDIN")
        self.cuckoo.shut_down(1)

    def outConnectionLost(self):
        logger.debug("Cuckoo closed STDOUT")
        self.cuckoo.shut_down(1)

    def errConnectionLost(self):
        logger.warning("Cuckoo closed STDERR")
        self.cuckoo.shut_down(1)

    def processExited(self, reason):
        logger.info("Cuckoo exited with status %s", reason.value.exitCode)
        self.cuckoo.shut_down()

    def processEnded(self, reason):
        logger.info("Cuckoo ended with status %s", reason.value.exitCode)
        self.cuckoo.shut_down()


class CuckooReport(object):
    """ Represents a Cuckoo analysis JSON report. """
    def __init__(self, report=None):
        """
        @param report: hash with report data from Cuckoo
        """
        if report is None:
            report = {}
        self.report = report

    @property
    def raw(self):
        return self.report

    @property
    def requested_domains(self):
        """
        Gets the requested domains from the Cuckoo report.

        @returns: The requested domains from the Cuckoo report.
        """
        try:
            return [d['request'] for d in self.report['network']['dns']]
        except KeyError:
            return []

    @property
    def signatures(self):
        """
        Gets the triggered signatures from the Cuckoo report.

        @returns: The triggered signatures from the Cuckoo report or None of
                  there was an error parsing the Cuckoo report.
        """
        return self.report.get('signatures', [])

    @property
    def signature_descriptions(self):
        """
        Gets the description of triggered Cuckoo signatures from report.

        @returns: The description of triggered signatures from the Cuckoo
                  report or empty list if there was an error parsing the
                  Cuckoo report.
        """
        descriptions = []
        for sig in self.signatures:
            descriptions.append(sig['description'])
        return descriptions

    @property
    def score(self):
        """
        Gets the score from the Cuckoo report.

        @returns: The score from the Cuckoo report or None of there was an
                  error parsing the Cuckoo report.
        """
        try:
            return self.report['info']['score']
        except KeyError:
            return 0.0

    @property
    def errors(self):
        """
        Errors occurred during Cuckoo analysis.

        @returns: The errors occurred during Cuckoo analysis or None of there
                  was an error parsing the Cuckoo report.
        """
        try:
            return self.report['debug']['errors']
        except KeyError:
            return []

    @property
    def cuckoo_server_messages(self):
        """
        Messages logged by the Cuckoo server (as opposed to those logged by the
        agent inside the analysis VM).

        @returns: List of messages.
        """
        if self.errors:
            logger.warning('Cuckoo produced %d error(s) during processing.' % len(self.errors))
        try:
            return self.report['debug']['cuckoo']
        except KeyError:
            return []



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# queuing.py                                                                  #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import logging
from threading import Thread, Event, Lock
from queue import Queue, Empty
from time import sleep
from peekaboo.ruleset import Result, RuleResult
from peekaboo.exceptions import PeekabooAnalysisDeferred, \
    PeekabooDatabaseError


logger = logging.getLogger(__name__)


class JobQueue:
    """ Peekaboo's queuing system. """
    def __init__(self, ruleset_engine, db_con, worker_count=4,
                 queue_timeout=300, shutdown_timeout=60,
                 cluster_duplicate_check_interval=5):
        """ Initialise job queue by creating n Peekaboo worker threads to
        process samples.

        @param db_con: Database connection object for cluster instance
                       coordination, i.e. saving sample info.
        @param worker_count: The amount of worker threads to create. Defaults to 4.
        @param queue_timeout: How long to block before considering queueing failed.
        """
        self.db_con = db_con
        self.jobs = Queue()
        self.workers = []
        self.worker_count = worker_count
        self.queue_timeout = queue_timeout
        self.shutdown_timeout = shutdown_timeout

        # keep a backlog of samples with hashes identical to samples currently
        # in analysis to avoid analysing multiple identical samples
        # simultaneously. Once one analysis has finished, we can submit the
        # others and the ruleset will notice that we already know the result.
        self.duplicates = {}
        self.duplock = Lock()

        # keep a similar backlog of samples currently being processed by
        # other instances so we can regularly try to resubmit them and re-use
        # the other instances' cached results from the database
        self.cluster_duplicates = {}

        for i in range(0, self.worker_count):
            logger.debug("Create Worker %d" % i)
            worker = Worker(i, self, ruleset_engine, db_con)
            self.workers.append(worker)
            worker.start()

        logger.info('Created %d Workers.' % self.worker_count)

        self.cluster_duplicate_handler = None
        if cluster_duplicate_check_interval:
            logger.debug("Starting cluster duplicate handler thread with "
                         "check interval %d.", cluster_duplicate_check_interval)
            self.cluster_duplicate_handler = ClusterDuplicateHandler(
                self, cluster_duplicate_check_interval)
            self.cluster_duplicate_handler.start();
        else:
            logger.debug("Disabling cluster duplicate handler thread.")

    def submit(self, sample, submitter):
        """
        Adds a Sample object to the job queue.
        If the queue is full, we block for 300 seconds and then throw an exception.

        @param sample: The Sample object to add to the queue.
        @param submitter: The name of the class / module that wants to submit the sample.
        @raises Full: if the queue is full.
        """
        sample_hash = sample.sha256sum
        sample_str = "%s" % sample
        duplicate = None
        cluster_duplicate = None
        resubmit = None
        # we have to lock this down because apart from callbacks from our
        # Workers we're also called from the ThreadingUnixStreamServer
        with self.duplock:
            # check if a sample with same hash is currently in flight
            duplicates = self.duplicates.get(sample_hash)
            if duplicates is not None:
                # we are regularly resubmitting samples, e.g. after we've
                # noticed that cuckoo is finished analysing them. This
                # obviously isn't a duplicate but continued processing of the
                # same sample.
                if duplicates['master'] == sample:
                    resubmit = sample_str
                    self.jobs.put(sample, True, self.queue_timeout)
                else:
                    # record the to-be-submitted sample as duplicate and do nothing
                    duplicate = sample_str
                    duplicates['duplicates'].append(sample)
            else:
                # are we the first of potentially multiple instances working on
                # this sample?
                try:
                    locked = self.db_con.mark_sample_in_flight(sample)
                except PeekabooDatabaseError as dberr:
                    logger.error(dberr)
                    return False

                if locked:
                    # initialise a per-duplicate backlog for this sample which
                    # also serves as in-flight marker and submit to queue
                    self.duplicates[sample_hash] = {
                            'master': sample,
                            'duplicates': [] }
                    self.jobs.put(sample, True, self.queue_timeout)
                else:
                    # another instance is working on this
                    if self.cluster_duplicates.get(sample_hash) is None:
                        self.cluster_duplicates[sample_hash] = []

                    cluster_duplicate = sample_str
                    self.cluster_duplicates[sample_hash].append(sample)

        if duplicate:
            logger.debug("Sample from %s is duplicate and waiting for "
                    "running analysis to finish: %s" % (submitter, duplicate))
        elif cluster_duplicate:
            logger.debug("Sample from %s is concurrently processed by "
                    "another instance and held: %s" % (submitter,
                        cluster_duplicate))
        elif resubmit:
            logger.debug("Resubmitted sample to job queue for %s: %s" %
                    (submitter, resubmit))
        else:
            logger.debug("New sample submitted to job queue by %s. %s" %
                    (submitter, sample_str))

        return True

    def submit_cluster_duplicates(self):
        if not self.cluster_duplicates.keys():
            return True

        submitted_cluster_duplicates = []

        with self.duplock:
            # try to submit *all* samples which have been marked as being
            # processed by another instance concurrently
            for sample_hash, sample_duplicates in self.cluster_duplicates.items():
                # try to mark as in-flight
                try:
                    locked = self.db_con.mark_sample_in_flight(
                        sample_duplicates[0])
                except PeekabooDatabaseError as dberr:
                    logger.error(dberr)
                    return False

                if locked:
                    sample_str = "%s" % sample_duplicates[0]
                    if self.duplicates.get(sample_hash) is not None:
                        logger.error("Possible backlog corruption for sample "
                                "%s! Please file a bug report. Trying to "
                                "continue..." % sample_str)
                        continue

                    # submit one of the held-back samples as a new master
                    # analysis in case the analysis on the other instance
                    # failed and we have no result in the database yet. If all
                    # is well, this master should finish analysis very quickly
                    # using the stored result, causing all the duplicates to be
                    # submitted and finish quickly as well.
                    sample = sample_duplicates.pop()
                    self.duplicates[sample_hash] = {
                            'master': sample,
                            'duplicates': sample_duplicates }
                    submitted_cluster_duplicates.append(sample_str)
                    self.jobs.put(sample, True, self.queue_timeout)
                    del self.cluster_duplicates[sample_hash]

        if len(submitted_cluster_duplicates) > 0:
            logger.debug("Submitted cluster duplicates (and potentially "
                    "their duplicates) from backlog: %s" %
                    submitted_cluster_duplicates)

        return True

    def clear_stale_in_flight_samples(self):
        try:
            cleared = self.db_con.clear_stale_in_flight_samples()
        except PeekabooDatabaseError as dberr:
            logger.error(dberr)
            cleared = False

        return cleared

    def submit_duplicates(self, sample_hash):
        """ Check if any samples have been held from processing as duplicates
        and submit them now. Clear the original sample whose duplicates have
        been submitted from the in-flight list.

        @param sample_hash: Hash of sample to check for duplicates
        """
        submitted_duplicates = []
        with self.duplock:
            # duplicates which have been submitted from the backlog still
            # report done but do not get registered as potentially having
            # duplicates because we expect the ruleset to identify them as
            # already known and process them quickly now that the first
            # instance has gone through full analysis. Therefore we can ignore
            # them here.
            if sample_hash not in self.duplicates:
                return

            # submit all samples which have accumulated in the backlog
            for s in self.duplicates[sample_hash]['duplicates']:
                submitted_duplicates.append("%s" % s)
                self.jobs.put(s, True, self.queue_timeout)

            sample = self.duplicates[sample_hash]['master']
            try:
                self.db_con.clear_sample_in_flight(sample)
            except PeekabooDatabaseError as dberr:
                logger.error(dberr)

            sample_str = "%s" % sample
            del self.duplicates[sample_hash]

        logger.debug("Cleared sample %s from in-flight list" % sample_str)
        if len(submitted_duplicates) > 0:
            logger.debug("Submitted duplicates from backlog: %s" % submitted_duplicates)

    def done(self, sample):
        """ Perform cleanup actions after sample processing is done:
        1. Submit held duplicates and
        2. notify request handler thread that sample processing is done.

        @param sample: The Sample object to post-process. """
        self.submit_duplicates(sample.sha256sum)

        # now that this sample is really done and cleared from the queue, tell
        # its connection handler about it
        sample.mark_done()

    def dequeue(self):
        """ Remove a sample from the queue. Used by the workers to get their
        work. Blocks indefinitely until some work is available. If we want to
        wake the workers for some other reason, we send them a None item as
        ping. """
        return self.jobs.get(True)

    def shut_down(self, timeout = None):
        if not timeout:
            timeout = self.shutdown_timeout

        logger.info("Shutting down. Giving workers %d seconds to stop" % timeout)

        if self.cluster_duplicate_handler:
            self.cluster_duplicate_handler.shut_down()

        # tell all workers to shut down
        for worker in self.workers:
            worker.shut_down()

        # put a ping for each worker on the queue. Since they already all know
        # that they're supposed to shut down, each of them will only remove
        # one item from the queue and then exit, leaving the others for their
        # colleagues. For this reason this loop can't be folded into the above!
        for worker in self.workers:
            self.jobs.put(None)

        # wait for workers to end
        interval = 1
        for attempt in range(1, timeout // interval + 1):
            still_running = []
            for worker in self.workers:
                if worker.running:
                    still_running.append(worker)

            self.workers = still_running
            if len(self.workers) == 0:
                break

            sleep(interval)
            logger.debug('%d: %d workers still running', attempt,
                         len(self.workers))

        if len(self.workers) > 0:
            logger.error("Some workers refused to stop.")

class ClusterDuplicateHandler(Thread):
    def __init__(self, job_queue, interval=5):
        self.shutdown_requested = Event()
        self.shutdown_requested.clear()
        self.job_queue = job_queue
        self.interval = interval
        Thread.__init__(self, name="ClusterDuplicateHandler")

    def run(self):
        logger.debug("Cluster duplicate handler started.")

        while not self.shutdown_requested.wait(self.interval):
            logger.debug("Checking for samples in processing by other "
                         "instances to submit")
            # TODO: Error handling: How do we cause Peekaboo to exit with an
            # error from here? For now just keep trying and hope (database)
            # failure is transient.
            self.job_queue.clear_stale_in_flight_samples()
            self.job_queue.submit_cluster_duplicates()

        logger.debug("Cluster duplicate handler shut down.")

    def shut_down(self):
        self.shutdown_requested.set()


class Worker(Thread):
    """ A Worker thread to process a sample. """
    def __init__(self, wid, job_queue, ruleset_engine, db_con):
        # whether we should run
        self.shutdown_requested = Event()
        self.shutdown_requested.clear()
        # whether we are actually running
        self.running_flag = Event()
        self.running_flag.clear()
        self.worker_id = wid
        self.job_queue = job_queue
        self.ruleset_engine = ruleset_engine
        self.db_con = db_con
        Thread.__init__(self, name="Worker-%d" % wid)

    def run(self):
        self.running_flag.set()
        while not self.shutdown_requested.is_set():
            logger.debug('Worker %d: Ready', self.worker_id)

            try:
                # wait blocking for next job (thread safe) with timeout
                sample = self.job_queue.dequeue()
            except Empty:
                continue

            if sample is None:
                # we just got pinged
                continue

            logger.info('Worker %d: Processing sample %s',
                        self.worker_id, sample)

            # The following used to be one big try/except block catching any
            # exception. This got complicated because in the case of
            # CuckooReportPending we use exceptions for control flow as well
            # (which might be questionable in itself). Instead of catching,
            # logging and ignoring errors here if workers start to die again
            # because of uncaught exceptions we should improve error handling
            # in the subroutines causing it.

            if not sample.init():
                logger.error('Sample initialization failed')
                sample.add_rule_result(
                    RuleResult(
                        "Worker", result=Result.failed,
                        reason=_("Sample initialization failed"),
                        further_analysis=False))
                self.job_queue.done(sample.sha256sum)
                continue

            try:
                self.ruleset_engine.run(sample)
            except PeekabooAnalysisDeferred:
                logger.debug("Report for sample %s still pending", sample)
                continue

            if sample.result >= Result.failed:
                sample.dump_processing_info()

            if sample.result != Result.failed:
                logger.debug('Saving results to database')
                try:
                    self.db_con.analysis_save(sample)
                except PeekabooDatabaseError as dberr:
                    logger.error('Failed to save analysis result to '
                                 'database: %s', dberr)
                    # no showstopper, we can limp on without caching in DB
            else:
                logger.debug('Not saving results of failed analysis')

            sample.cleanup()
            self.job_queue.done(sample)

        logger.info('Worker %d: Stopped' % self.worker_id)
        self.running_flag.clear()

    def shut_down(self):
        self.shutdown_requested.set()

    @property
    def running(self):
        return self.running_flag.is_set()



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# toolbox/                                                                    #
#         peekabooyar.py                                                      #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


from peekaboo.ruleset import Result
from peekaboo.ruleset.rules import Rule
import yara


class ContainsPeekabooYarRule(Rule):
    """
    Checks the given sample for the PeekabooYar (EICAR like) malicious string.
    """
    rule_name = 'contains_peekabooyar'

    def evaluate(self, s):
        rules = yara.compile(
            source='''
            rule peekabooyar
            {
                strings:
                    $peekabooyar1 = "X5O!P%@AP-/_(:)_/-X22x8cz2$PeekabooAV-STD-ANTIVIRUS-TEST-FILE!$H+H*"

                condition:
                    $peekabooyar1
            }'''
        )

        # FIXME: Only user of file_path. Remove?
        with open(s.file_path, 'rb') as sample_file:
            matches = rules.match(data=sample_file.read())

        if matches != []:
            return self.result(Result.bad,
                               "Die Datei beinhaltet Peekabooyar",
                               False)

        return self.result(Result.unknown,
                           "Die Datei beinhaltet kein erkennbares Peekabooyar",
                           True)



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# sample.py                                                                   #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


import os
import hashlib
import json
import random
import re
import shutil
import string
import logging
import tempfile
# python 3's open with encoding parameter and implicit usage of the system
# locale-specified encoding
from builtins import open
from datetime import datetime
from peekaboo.ruleset import Result


logger = logging.getLogger(__name__)

class SampleFactory(object):
    """ A class for churning out loads of mostly identical sample objects.
    Contains all the global configuration data and object references each
    sample needs and thus serves as a registry of potential API breakage
    perhaps deserving looking into. """
    def __init__(self, cuckoo, base_dir, job_hash_regex,
                 keep_mail_data, processing_info_dir):
        # object references for interaction
        self.cuckoo = cuckoo

        # configuration
        self.base_dir = base_dir
        self.job_hash_regex = job_hash_regex
        self.keep_mail_data = keep_mail_data
        self.processing_info_dir = processing_info_dir

    def make_sample(self, file_path, status_change=None, metainfo=None):
        """ Create a new Sample object based on the factory's configured
        defaults and variable parameters. """
        return Sample(file_path, self.cuckoo, status_change, metainfo,
                      self.base_dir, self.job_hash_regex, self.keep_mail_data,
                      self.processing_info_dir)


class Sample(object):
    """
    This class handles and describes samples to be analysed by Peekaboo.

    A sample has properties like filename, MIME type, checksum or file size.
    These are accessible as properties. Most properties determine their value
    on first access, especially if that determination is somewhat expensive
    such as the file checksum.

    The data structure works together with Cuckoo to run behavioral analysis.
    """
    def __init__(self, file_path, cuckoo=None, status_change=None,
                 metainfo=None, base_dir=None, job_hash_regex=None,
                 keep_mail_data=False, processing_info_dir=None):
        self.__path = file_path
        self.__cuckoo = cuckoo
        self.__wd = None
        self.__filename = os.path.basename(self.__path)
        # A symlink that points to the actual file named
        # sha256sum.suffix
        self.__submit_path = None
        self.__cuckoo_job_id = -1
        self.__cuckoo_failed = False
        self.__cuckoo_report = None
        self.__oletools_report = None
        self.__filetools_report = None
        self.__done = False
        self.__status_change = status_change
        self.__result = Result.unchecked
        self.__reason = None
        self.__report = []  # Peekaboo's own report
        self.__internal_report = []
        self.__file_stat = None
        self.__sha256sum = None
        self.__file_extension = None
        self.__base_dir = base_dir
        self.__job_hash = None
        self.__job_hash_regex = job_hash_regex
        self.__keep_mail_data = keep_mail_data
        self.__processing_info_dir = processing_info_dir

        # Additional attributes for a sample object (i.e. meta info)
        # We do not make these private for the following reasons:
        # - this way they still somewhat resemble the previous arbitrary
        #   attribute dictionary idea
        # - we'd have to implement the name mangling for setting below
        # Even though, it is not recommended to access them directly they're an
        # implementation detail. We add respective properties for that.
        #
        # Security: Add more below to allow them to be accepted from the
        # client. We don't want anyone to be able to pollute our sample
        # objects from the outside. This also serves as a registry of what we
        # actually use and know how to deal with.
        self.meta_info_name_declared = None
        self.meta_info_type_declared = None
        self.meta_info_content_disposition = None

        self.initialized = False

        if metainfo:
            member_variables = vars(self)
            for field in metainfo:
                logger.debug('meta_info_%s = %s', field, metainfo[field])

                # JSON will transfer null/None values but we don't want them as
                # attributes in that case
                member = 'meta_info_%s' % field
                if member in member_variables and metainfo[field] is not None:
                    member_variables[member] = metainfo[field]

    def init(self):
        """
        Initialize the Sample object.

        The actual initialization is done here, because the main thread should
        not do the heavy lifting of e. g. parsing the meta info file to be able
        to accept new connections as quickly as possible.
        Instead, it only adds the sample objects to the queue and the workers
        to the actual initialization.
        """
        if self.initialized:
            return True

        logger.debug("initializing sample")

        # create a temporary directory where mkdtemp makes sure that
        # creation is atomic, i.e. no other process is using it
        try:
            self.__wd = tempfile.mkdtemp(
                prefix=self.job_hash, dir=self.__base_dir)
        except OSError as oserr:
            logger.error('Error creating working directory: %s', oserr)
            return False

        logger.debug('Working directory %s created', self.__wd)

        # create a symlink to submit the file with the correct file extension
        # to cuckoo via submit.py. This is so we do not leak the original
        # filename by default.
        submit_name = self.sha256sum
        if self.file_extension:
            submit_name = '%s.%s' % (submit_name, self.file_extension)

        self.__submit_path = os.path.join(self.__wd, submit_name)

        try:
            os.symlink(self.__path, self.__submit_path)
        except OSError as oserr:
            logger.error('Error linking sample from %s to working '
                         'directory as %s',
                         self.__path, self.__submit_path)
            self.cleanup()
            return False

        logger.debug('Sample symlinked from %s to %s',
                     self.__path, self.__submit_path)

        self.initialized = True

        self.__report.append(_("File \"%s\" %s is being analyzed")
                             % (self.__filename, self.sha256sum))

        # log some additional info to report to aid debugging
        if self.meta_info_name_declared:
            self.__internal_report.append("meta info: name_declared: %s"
                                          % self.meta_info_name_declared)

        if self.meta_info_type_declared:
            self.__internal_report.append("meta info: type_declared: %s"
                                          % self.meta_info_type_declared)

        return True

    @property
    def file_path(self):
        """ Returns the path to the sample given on creation including
        directories and filename. """
        return self.__path

    @property
    def filename(self):
        """ Returns the name of the sample file, i.e. the basename without path
        but including the file extension. """
        return self.__filename

    @property
    def result(self):
        """ Returns the overall evaluation result for this sample.

        @returns: peekaboo.ruleset.Result """
        return self.__result

    @property
    def reason(self):
        """ Gets the reason given by the rule determining the result which
        ended up as the overall evaluation result of this sample.

        @returns: string """
        return self.__reason

    @property
    def peekaboo_report(self):
        """ Return Peekaboo's report meant for the client, detailing what's
        been found on this sample.

        @returns: List of strings.
        """
        # This message used to be:
        # "Die Datei \"%s\" wurde als \"%s\" eingestuft\n\n"
        # Changed intentionally to not trigger configured god/bad matching
        # patterns in clients (e.g. AMaViS) any more since we switched to
        # reporting an overall analysis batch result.
        return self.__report + [_("File \"%s\" is considered \"%s\"")
                                % (self.__filename, self.__result.name)]

    @property
    def done(self):
        """ Tells whether the analysis of the sample is done, i.e. a final
        verdict has been reached and a result and reason are available. """
        return self.__done

    def mark_done(self):
        """ Mark this sample as done, i.e. fully analysed and verdict reached.
        """
        self.__done = True
        if self.__status_change:
            # notify whoever is interested that our status has changed
            self.__status_change.set()

    def generate_job_hash(self, size=8):
        """
        Generates a job hash (default: 8 characters).

        @param size: The amount of random characters to use for a job hash.
                     Defaults to 8.
        @return: a job hash consisting of a static prefix, a timestamp
                 representing the time when the method was invoked, and random
                 characters.
        """
        job_hash = 'peekaboo-run_analysis-'
        job_hash += '%s-' % datetime.now().strftime('%Y%m%dT%H%M%S')
        job_hash += ''.join(
            random.choice(string.digits + string.ascii_lowercase
                          + string.ascii_uppercase) for _ in range(size))
        return job_hash

    @property
    def job_hash(self):
        """ Returns a job identifier extracted from the file path using a
        configurable regular expression for use in other temporary or permanent
        (dump) path names to keep correlation to the original input job. """
        if self.__job_hash:
            return self.__job_hash

        match = re.search(self.__job_hash_regex, self.__path)
        if match is not None:
            job_hash = match.group(1)
        else:
            # regex did not match.
            # so we generate our own job hash and create the
            # working directory.
            job_hash = self.generate_job_hash()

        logger.debug("Job hash for this sample: %s" % job_hash)
        self.__job_hash = job_hash
        return job_hash

    def add_rule_result(self, res):
        """ Add a rule result to the sample. This also adds a message about
        this to the report and updates the overall analysis result (so far).
        """
        logger.debug('Adding rule result %s', res)
        self.__report.append(_("File \"%s\": %s") % (self.__filename, res))

        logger.debug("Current overall result: %s, new rule result: %s",
                     self.__result, res.result)
        # check if result of this rule is worse than what we know so far
        if res.result >= self.__result:
            self.__result = res.result
            self.__reason = res.reason

    def dump_processing_info(self):
        """
        Saves the Cuckoo report as HTML + JSON
        to a directory named after the job hash.
        """
        if not self.__processing_info_dir:
            logger.debug('Not dumping processing info because no path for the '
                         'data is unconfigured.')
            return

        dump_dir = os.path.join(self.__processing_info_dir, self.job_hash)
        if not os.path.isdir(dump_dir):
            try:
                os.makedirs(dump_dir, 0o770)
            except OSError as oserr:
                logger.error('Failed to create dump directory %s: %s',
                             dump_dir, oserr)
                return

        filename = self.__filename + '-' + self.sha256sum

        logger.debug('Dumping processing info to %s for sample %s',
                     dump_dir, self)

        # Peekaboo's report
        peekaboo_report = os.path.join(dump_dir, filename + '_report.txt')
        try:
            with open(peekaboo_report, 'w+') as pr_file:
                if self.__report:
                    pr_file.write('\n'.join(self.__report + [""]))
                if self.__internal_report:
                    pr_file.write('\n'.join(self.__internal_report + [""]))
        except (OSError, IOError) as error:
            logger.error('Failure to write report file %s: %s',
                         peekaboo_report, error)
            return

        # store malicious sample along with the reports
        if self.__result == Result.bad:
            try:
                shutil.copyfile(
                    self.__path,
                    os.path.join(dump_dir, self.__filename)
                )
            except (shutil.Error, IOError, OSError) as error:
                logger.error('Failure to copy sample file %s to dump '
                             'directory: %s', self.__path, error)
                return

        # Cuckoo report
        if self.__cuckoo_report:
            cuckoo_report = os.path.join(dump_dir,
                                         filename + '_cuckoo_report.json')
            try:
                with open(cuckoo_report, 'wb+') as cr_json_file:
                    cr_json = json.dumps(self.__cuckoo_report.raw,
                                         indent=1, ensure_ascii=True)
                    cr_json_file.write(cr_json.encode('ascii'))
            except (OSError, IOError) as error:
                logger.error('Failure to dump json report to %s: %s',
                             cuckoo_report, error)
                return

    @property
    def sha256sum(self):
        """ Returns the SHA256 checksum/fingerprint of this sample. Determines
        it automatically on first call. """

        if not self.__sha256sum:
            with open(self.__path, 'rb') as f:
                checksum = hashlib.sha256(f.read()).hexdigest()
                self.__sha256sum = checksum

        return self.__sha256sum

    @property
    def content_disposition(self):
        """ Returns the content disposition in the original email, e.g. inline
        or attachment, None if not available. """
        return self.meta_info_content_disposition

    @property
    def name_declared(self):
        """ Returns the name declared by the sample as its original filename,
        None if not available. """
        return self.meta_info_name_declared

    @property
    def file_extension(self):
        """ Determines the file extension of this sample. """
        if self.__file_extension:
            return self.__file_extension

        # try to find a file name containing an extension. Using
        # self.__filename will almost never yield anything useful because
        # amavis intentionally hands us files named only p001, p002 and so on.
        # But we still try it in case there's no declared name.
        filename = self.__filename
        if self.name_declared:
            filename = self.name_declared

        # extension or the empty string if none found
        self.__file_extension = os.path.splitext(filename)[1][1:]
        return self.__file_extension

    @property
    def type_declared(self):
        """ Returns the MIME type declared by the original MIME part, None if
        not available. """
        return self.meta_info_type_declared

    @property
    def job_id(self):
        return self.__cuckoo_job_id

    @property
    def file_size(self):
        """ Determine and cache sample file size

        @raises: OSError if e.g. file does not exist or is inaccessible """
        if not self.__file_stat:
            self.__file_stat = os.stat(self.__path)

        return self.__file_stat.st_size

    @property
    def cuckoo_failed(self):
        """ Returns whether the Cuckoo analysis failed. """
        return self.__cuckoo_failed

    @property
    def cuckoo_report(self):
        """ Returns the cuckoo report """
        return self.__cuckoo_report

    @property
    def oletools_report(self):
        """ Returns the oletools report """
        return self.__oletools_report

    @property
    def filetools_report(self):
        """ Returns the filetools report """
        return self.__filetools_report

    @property
    def submit_path(self):
        """ Returns the path to use for submission to Cuckoo """
        return self.__submit_path

    def submit_to_cuckoo(self):
        """ Submit the sample to Cuckoo for analysis and record job id.

        @raises: CuckooAnalsisFailedException if submission failed
        @returns: cuckoo job id
        """
        logger.debug("Submitting %s to Cuckoo", self.__submit_path)
        self.__cuckoo_job_id = self.__cuckoo.submit(self)
        self.__internal_report.append(
            _('Sample %s successfully submitted to Cuckoo as job %d')
            % (self, self.__cuckoo_job_id))
        return self.__cuckoo_job_id

    def mark_cuckoo_failure(self):
        """ Records whether Cuckoo analysis failed. """
        self.__cuckoo_failed = True

    def register_cuckoo_report(self, report):
        """ Records a Cuckoo report for later evaluation. """
        self.__cuckoo_report = report

    def register_oletools_report(self, report):
        """ Records a Oletools report for alter evaluation. """
        self.__oletools_report = report

    def register_filetools_report(self, report):
        """ Records a Filetools report for alter evaluation. """
        self.__filetools_report = report

    def cleanup(self):
        """ Clean up after the sample has been analysed, removing a potentially
        created workdir. """
        # nothing to do if we never created a workdir
        if not self.__wd:
            return

        if self.__keep_mail_data:
            logger.debug('Keeping mail data in %s', self.__wd)
            return

        logger.debug("Deleting working directory %s", self.__wd)
        try:
            shutil.rmtree(self.__wd)
        except OSError as oserr:
            logger.error('Failed to remove working directory %s: %s',
                         self.__wd, oserr)

    def __str__(self):
        return ("<Sample(filename='%s', job_id='%d',"
                " result='%s', sha256sum='%s')>"
                % (self.__filename,
                   self.__cuckoo_job_id,
                   self.__result,
                   self.sha256sum))

    __repr__ = __str__



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# exceptions.py                                                               #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


class PeekabooException(Exception):
    """ General exception class for all custom exception classes of Peekaboo. """
    pass


class PeekabooConfigException(PeekabooException):
    pass


class PeekabooDatabaseError(PeekabooException):
    pass


class PeekabooRulesetException(PeekabooException):
    pass


class PeekabooRulesetConfigError(PeekabooException):
    """ Used to signal that a rule is unhappy with its configuration. """
    pass


class PeekabooAnalysisDeferred(PeekabooRulesetException):
    """ Analysis has been deferred to a later point in time.

    An exception signifying that analysis has been deferred to a later point in
    time. Ruleset processing will be aborted (without error). Useful if we're
    waiting for someone to finish their analysis and defer our interpretation
    of their findings until they become available, most notably the Cuckoo
    report.

    Not an exception in the traditional sense since it does not indicate an
    error but actually influences control flow instead. Somewhat questionable
    in that regard.

    The raiser becomes owner of the sample and is responsible to appropriately
    resubmit it into Peekaboo once it wants processing to continue. That should
    take into account that the ruleset will be rerun from the very beginning.
    """
    pass


class CuckooSubmitFailedException(PeekabooException):
    """ An exception raised if submitting a job to Cuckoo fails. """
    pass



#!/usr/bin/env python
# encoding: utf-8

###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# peekaboo-util.py                                                            #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


from __future__ import print_function
from os import path
from argparse import ArgumentParser
import json
import socket
import re
import logging
import sys


logging.basicConfig()
logger = logging.getLogger(__name__)


class PeekabooUtil(object):
    """ Utility fo interface with Peekaboo API over the socket connection """
    def __init__(self, socket_file):
        logger.debug('Initialising PeekabooUtil')
        self.peekaboo = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        logger.debug('Opening socket %s', socket_file)
        self.peekaboo.connect(socket_file)

    def send_receive(self, request):
        """ Send request to peekaboo and return its answer """
        logger.debug('Sending request: %s', request)

        self.peekaboo.send(request.encode('utf-8'))
        logger.info('Waiting for response...')

        buf = ''
        while True:
            data = self.peekaboo.recv(1024)
            if data:
                buf += data.decode('utf-8')
            else:
                self.peekaboo.close()
                break
        logger.debug('Received from peekaboo: %s', buf)
        return buf

    def send_receive_json(self, data):
        """ Send and receive data in JSON format. """
        request = json.dumps(data)
        response = self.send_receive(request)
        outdata = None
        for line in response.splitlines():
            try:
                # try to parse and stop at first thing that parses
                outdata = json.loads(line)
                break
            except ValueError as error:
                # FIXME: daemon talks a mix of plain text and JSON. So for now
                # we must ignore everything that doesn't parse. This includes
                # errors.  We can't even employ a heuristic since all of those
                # can be translated.
                pass

        return outdata

    def ping(self):
        """ Send ping request to daemon and optionally print response. """
        logger.debug("Sending ping...")
        try:
            pong = self.send_receive_json([{"request": "ping"}])
        except socket.error as error:
            logger.error("Error communicating with daemon: %s", error)
            return 2

        if not isinstance(pong, dict):
            logger.error("Invalid response from daemon: %s", pong)
            return 2

        reqtype = pong.get('request')
        response = pong.get('response')
        if reqtype is None or response is None:
            logger.error("Incomplete response from daemon: %s", pong)
            return 2

        if reqtype != 'ping' and response != 'pong':
            logger.error("Response is not a pong")
            return 2

        logger.info('Pong received.')
        return 0

    def raw(self, raw):
        """ Send raw data to the daemon and display response. """
        logger.debug("Sending raw...")
        try:
            response = self.send_receive(raw)
        except socket.error as error:
            logger.error("Error communicating with daemon: %s", error)
            return 2

        print(response)
        return 0

    def scan_file(self, filenames):
        """ Scan the supplied filenames with peekaboo and output result """
        result_regex = re.compile(r'has been categorized',
                                  re.MULTILINE + re.DOTALL + re.UNICODE)
        requests = []
        for filename in filenames:
            requests.append({"request": "scan-file",
                             "full_name": path.abspath(filename)})

        buf = self.send_receive(json.dumps(requests))

        exit_code = 0
        for result in buf.splitlines():
            output = result_regex.search(result)
            if output:
                if 'bad' in result:
                    print(result)
                    exit_code = 1
                logger.info(result)

        return exit_code

def main():
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(help='commands')

    parser.add_argument('-v', '--verbose', action='store_true', required=False,
                        help='List results of all files not only bad ones')
    parser.add_argument('-vv', '--verbose2', action='store_true', required=False,
                        help='List detailed analysis results of every rule')
    parser.add_argument('-d', '--debug', action='store_true', required=False,
                        help='Output additional diagnostics')
    parser.add_argument('-t', '--timeout', type=float, required=False,
                        default=None, help='Communications timeout')
    parser.add_argument('-s', '--socket-file', action='store', required=True,
                        help='Path to Peekaboo\'s socket file')

    scan_file_parser = subparsers.add_parser('scan-file',
                                             help='Scan a file and report it')
    scan_file_parser.add_argument('-f', '--filename', action='append', required=True,
                                  help='Path to the file to scan. Can be given more '
                                       'than once to scan multiple files.')
    scan_file_parser.set_defaults(func=command_scan_file)

    ping_parser = subparsers.add_parser('ping', help='Ping the daemon')
    ping_parser.set_defaults(func=command_ping)

    raw_parser = subparsers.add_parser('raw',
                                       help='Send raw input to the daemon')
    raw_parser.add_argument('-j', '--json', action='store', required=True,
                            help='Raw JSON to send to daemon')
    raw_parser.set_defaults(func=command_raw)

    args = parser.parse_args()

    logger.setLevel(logging.ERROR)
    if args.verbose:
        logger.setLevel(logging.INFO)
    if args.verbose2 or args.debug:
        logger.setLevel(logging.DEBUG)

    if args.timeout:
        socket.setdefaulttimeout(args.timeout)

    try:
        util = PeekabooUtil(args.socket_file)
    except socket.error as error:
        logger.error("Error connecting to peekaboo socket: %s", error)
        return 2

    return args.func(util, args)


def command_scan_file(util, args):
    """ Handler for command scan_file """
    return util.scan_file(args.filename)


def command_ping(util, _):
    """ Handler for command ping """
    return util.ping()


def command_raw(util, args):
    """ Handler for command raw """
    return util.raw(args.json)


if __name__ == "__main__":
    sys.exit(main())



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# config.py                                                                   #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

""" The configuration for the main program as well as the ruleset. Handles
defaults as well as reading a configuration file. """


import re
import sys
import logging
import configparser
from peekaboo.exceptions import PeekabooConfigException


logger = logging.getLogger(__name__)

class PeekabooConfigParser( # pylint: disable=too-many-ancestors
        configparser.ConfigParser):
    """ A config parser that gives error feedback if a required file does not
    exist or cannot be opened. """
    LOG_LEVEL = object()
    OCTAL = object()
    RELIST = object()
    IRELIST = object()

    def __init__(self, config_file):
        # super() does not work here because ConfigParser uses old-style
        # classes in python 2
        configparser.ConfigParser.__init__(self)

        try:
            self.read_file(open(config_file))
        except IOError as ioerror:
            raise PeekabooConfigException(
                'Configuration file "%s" can not be opened for reading: %s' %
                (config_file, ioerror))
        except configparser.Error as cperror:
            raise PeekabooConfigException(
                'Configuration file "%s" can not be parsed: %s' %
                (config_file, cperror))

        self.lists = {}
        self.relists = {}

    def getlist(self, section, option, raw=False, vars=None, fallback=None):
        """ Special getter where multiple options in the config file
        distinguished by a .<no> suffix form a list. Matches the signature for
        configparser getters. """
        # cache results because the following is somewhat inefficient
        if section not in self.lists:
            self.lists[section] = {}

        if option in self.lists[section]:
            return self.lists[section][option]

        if section not in self:
            self.lists[section][option] = fallback
            return fallback

        # Go over all options in this section we want to allow "holes" in
        # the lists, i.e setting.1, setting.2 but no setting.3 followed by
        # setting.4. We use here that ConfigParser retains option order from
        # the file.
        value = []
        for setting in self[section]:
            if not setting.startswith(option):
                continue

            # Parse 'setting' into (key) and 'setting.subscript' into
            # (key, subscript) and use it to determine if this setting is a
            # list. Note how we do not use the subscript at all here.
            name_parts = setting.split('.')
            key = name_parts[0]
            is_list = len(name_parts) > 1

            if key != option:
                continue

            if not is_list:
                raise PeekabooConfigException(
                    'Option %s in section %s is supposed to be a list '
                    'but given as individual setting' % (setting, section))

            # Potential further checks:
            # - There are no duplicate settings with ConfigParser. The last
            #   one always wins.

            value.append(self[section].get(setting, raw=raw, vars=vars))

        # it's not gonna get any better on the next call, so cache even the
        # default
        if not value:
            value = fallback

        self.lists[section][option] = value
        return value

    def getirelist(self, section, option, raw=False, vars=None, fallback=None, flags=None):
        """ Special getter for lists of regular expressions that are compiled to match
        case insesitive (IGNORECASE). Returns the compiled expression objects in a
        list ready for matching and searching.
        """
        return self.getrelist(section, option, raw=raw, vars=vars, fallback=fallback, flags=re.IGNORECASE)

    def getrelist(self, section, option, raw=False, vars=None, fallback=None, flags=0):
        """ Special getter for lists of regular expressions. Returns the
        compiled expression objects in a list ready for matching and searching.
        """
        if section not in self.relists:
            self.relists[section] = {}

        if option in self.relists[section]:
            return self.relists[section][option]

        if section not in self:
            self.relists[section][option] = fallback
            return fallback

        strlist = self[section].getlist(option, raw=raw, vars=vars,
                                        fallback=fallback)
        if strlist is None:
            self.relists[section][option] = None
            return None

        compiled_res = []
        for regex in strlist:
            try:
                compiled_res.append(re.compile(regex, flags))
            except (ValueError, TypeError) as error:
                raise PeekabooConfigException(
                    'Failed to compile regular expression "%s" (section %s, '
                    'option %s): %s' % (re, section, option, error))

        # it's not gonna get any better on the next call, so cache even the
        # default
        if not compiled_res:
            compiled_res = fallback

        self.relists[section][option] = compiled_res
        return compiled_res

    def get_log_level(self, section, option, raw=False, vars=None,
                      fallback=None):
        """ Get the log level from the configuration file and parse the string
        into a logging loglevel such as logging.CRITICAL. Raises config
        exception if the log level is unknown. Options identical to get(). """
        levels = {
            'CRITICAL': logging.CRITICAL,
            'ERROR': logging.ERROR,
            'WARNING': logging.WARNING,
            'INFO': logging.INFO,
            'DEBUG': logging.DEBUG
        }

        level = self.get(section, option, raw=raw, vars=vars, fallback=None)
        if level is None:
            return fallback

        if level not in levels:
            raise PeekabooConfigException('Unknown log level %s' % level)

        return levels[level]

    def getoctal(self, section, option, raw=False, vars=None, fallback=None):
        """ Get an integer in octal notation. Raises config
        exception if the format is wrong. Options identical to get(). """
        value = self.get(section, option, raw=raw, vars=vars, fallback=None)
        if value is None:
            return fallback

        try:
            octal = int(value, 8)
        except ValueError:
            raise PeekabooConfigException(
                'Invalid value for octal option %s in section %s: %s'
                % (option, section, value))

        return octal

    def get_by_type(self, section, option, fallback=None, option_type=None):
        """ Get an option from the configuration file parser. Automatically
        detects the type from the type of the default if given and calls the
        right getter method to coerce the value to the correct type.

        @param section: Which section to look for option in.
        @type section: string
        @param option: The option to read.
        @type option: string
        @param fallback: (optional) Default value to return if option is not
                         found. Defaults itself to None so that the method will
                         return None if the option is not found.
        @type fallback: int, bool, str or None.
        @param option_type: Override the option type.
        @type option_type: int, bool, str or None. """
        if option_type is None and fallback is not None:
            option_type = type(fallback)

        getter = {
            int: self.getint,
            float: self.getfloat,
            bool: self.getboolean,
            list: self.getlist,
            tuple: self.getlist,
            str: self.get,
            None: self.get,

            # these only work when given explicitly as option_type
            self.LOG_LEVEL: self.get_log_level,
            self.OCTAL: self.getoctal,
            self.RELIST: self.getrelist,
            self.IRELIST: self.getirelist,
        }

        return getter[option_type](section, option, fallback=fallback)

    def check_config(self, known_options):
        """ Check this configuration against a list of known options. Raise an
        exception if any unknown options are found.

        @param known_options: A dict of sections and options, the key being the
                              section name and the value a list of option names.
        @type known_options: dict

        @returns: None
        @raises PeekabooConfigException: if any unknown sections or options are
                                         found.
        """
        known_sections = known_options.keys()
        self.check_sections(known_sections)

        # go over sections both allowed and in the config
        for section in known_sections:
            self.check_section_options(section, known_options[section])

    def check_sections(self, known_sections):
        """ Check a list of known section names against this configuration

        @param known_sections: names of known sections
        @type known_sections: list(string)

        @returns: None
        @raises PeekabooConfigException: if any unknown sections are found in
                                         the configuration.
        """
        section_diff = set(self.sections()) - set(known_sections)
        if section_diff:
            raise PeekabooConfigException(
                'Unknown section(s) found in config: %s'
                % ', '.join(section_diff))

    def check_section_options(self, section, known_options):
        """ Check a config section for unknown options.

        @param section: name of section to check
        @type section: string
        @param known_options: list of names of known options to check against
        @type known_options: list(string)

        @returns: None
        @raises PeekabooConfigException: if any unknown options are found. """
        try:
            section_options = map(
                # account for option.1 list syntax
                lambda x: x.split('.')[0],
                self.options(section))
        except configparser.NoSectionError:
            # a non-existant section can have no non-allowed options :)
            return

        option_diff = set(section_options) - set(known_options)
        if option_diff:
            raise PeekabooConfigException(
                'Unknown config option(s) found in section %s: %s'
                % (section, ', '.join(option_diff)))


class PeekabooConfig(PeekabooConfigParser):
    """ This class represents the Peekaboo configuration. """
    def __init__(self, config_file=None, log_level=None):
        """ Initialise the configuration with defaults, overwrite with command
        line options and finally read the configuration file. """
        # hard defaults: The idea here is that every config option has a
        # default that would in principle enable Peekaboo to run. Code using
        # the option should still cope with no or an empty value being handed
        # to it.
        self.user = 'peekaboo'
        self.group = None
        self.pid_file = '/var/run/peekaboo/peekaboo.pid'
        self.sock_file = '/var/run/peekaboo/peekaboo.sock'
        self.sock_group = None
        self.sock_mode = 0o0660
        self.log_level = logging.INFO
        self.log_format = '%(asctime)s - %(name)s - (%(threadName)s) - ' \
                          '%(levelname)s - %(message)s'
        self.interpreter = '/usr/bin/python2 -u'
        self.worker_count = 3
        self.sample_base_dir = '/tmp'
        self.job_hash_regex = '/amavis/tmp/([^/]+)/parts/'
        self.keep_mail_data = False
        self.processing_info_dir = '/var/lib/peekaboo/malware_reports'
        self.report_locale = None
        self.db_url = 'sqlite:////var/lib/peekaboo/peekaboo.db'
        self.db_log_level = logging.WARNING
        self.config_file = '/opt/peekaboo/etc/peekaboo.conf'
        self.ruleset_config = '/opt/peekaboo/etc/ruleset.conf'
        self.cuckoo_mode = "api"
        self.cuckoo_url = 'http://127.0.0.1:8090'
        self.cuckoo_api_token = ''
        self.cuckoo_poll_interval = 5
        self.cuckoo_submit_original_filename = True
        self.cuckoo_maximum_job_age = 15*60
        self.cuckoo_storage = '/var/lib/peekaboo/.cuckoo/storage'
        self.cuckoo_exec = '/opt/cuckoo/bin/cuckoo'
        self.cuckoo_submit = '/opt/cuckoo/bin/cuckoo submit'
        self.cluster_instance_id = 0
        self.cluster_stale_in_flight_threshold = 15*60
        self.cluster_duplicate_check_interval = 60

        # section and option names for the configuration file. key is the above
        # variable name whose value will be overwritten by the configuration
        # file value. Third item can be getter function if special parsing is
        # required.
        config_options = {
            'log_level': ['logging', 'log_level', self.LOG_LEVEL],
            'log_format': ['logging', 'log_format'],
            'user': ['global', 'user'],
            'group': ['global', 'group'],
            'pid_file': ['global', 'pid_file'],
            'sock_file': ['global', 'socket_file'],
            'sock_group': ['global', 'socket_group'],
            'sock_mode': ['global', 'socket_mode', self.OCTAL],
            'interpreter': ['global', 'interpreter'],
            'worker_count': ['global', 'worker_count'],
            'sample_base_dir': ['global', 'sample_base_dir'],
            'job_hash_regex': ['global', 'job_hash_regex'],
            'keep_mail_data': ['global', 'keep_mail_data'],
            'processing_info_dir': ['global', 'processing_info_dir'],
            'report_locale': ['global', 'report_locale'],
            'db_url': ['db', 'url'],
            'db_log_level': ['db', 'log_level', self.LOG_LEVEL],
            'ruleset_config': ['ruleset', 'config'],
            'cuckoo_mode': ['cuckoo', 'mode'],
            'cuckoo_url': ['cuckoo', 'url'],
            'cuckoo_api_token': ['cuckoo', 'api_token'],
            'cuckoo_poll_interval': ['cuckoo', 'poll_interval'],
            'cuckoo_submit_original_filename': [
                'cuckoo', 'submit_original_filename'],
            'cuckoo_maximum_job_age': ['cuckoo', 'maximum_job_age'],
            'cuckoo_storage': ['cuckoo', 'storage_path'],
            'cuckoo_exec': ['cuckoo', 'exec'],
            'cuckoo_submit': ['cuckoo', 'submit'],
            'cluster_instance_id': ['cluster', 'instance_id'],
            'cluster_stale_in_flight_threshold': ['cluster', 'stale_in_flight_threshold'],
            'cluster_duplicate_check_interval': ['cluster', 'duplicate_check_interval'],
        }

        # overrides from outside, e.g. by command line arguments whose values
        # are needed while reading the configuration file already (most notably
        # log level and path to the config file).
        if log_level:
            self.log_level = log_level
        if config_file:
            self.config_file = config_file

        # setup default logging to log any errors during the
        # parsing of the config file.
        self.setup_logging()

        # read configuration file. Note that we require a configuration file
        # here. We may change that if we decide that we want to allow the user
        # to run us with the above defaults only.
        PeekabooConfigParser.__init__(self, self.config_file)

        # overwrite above defaults in our member variables via indirect access
        settings = vars(self)
        check_options = {}
        for (setting, config_names) in config_options.items():
            section = config_names[0]
            option = config_names[1]

            # remember for later checking for unknown options
            if section not in check_options:
                check_options[section] = []
            check_options[section].append(option)

            # maybe force the option's value type
            option_type = None
            if len(config_names) == 3:
                option_type = config_names[2]

            # e.g.:
            # self.log_format = self.get('logging', 'log_format',
            #                            self.log_format)
            settings[setting] = self.get_by_type(
                section, option, fallback=settings[setting],
                option_type=option_type)

        # now check for unknown options
        self.check_config(check_options)

        # Update logging with what we just parsed from the config
        self.setup_logging()

        # here we could overwrite defaults and config file with additional
        # command line arguments if required

    def setup_logging(self):
        """ Setup logging to console by reconfiguring the root logger so that
        it affects all loggers everywhere.  """
        _logger = logging.getLogger()

        # Check if we already have a log handler
        if _logger.handlers:
            # Remove all handlers
            for handler in _logger.handlers:
                _logger.removeHandler(handler)
        # log format
        log_formatter = logging.Formatter(self.log_format)
        # create console handler and set level to debug
        to_console_log_handler = logging.StreamHandler(sys.stdout)
        to_console_log_handler.setFormatter(log_formatter)
        _logger.addHandler(to_console_log_handler)
        _logger.setLevel(self.log_level)

    def __str__(self):
        settings = {}
        for (option, value) in vars(self).items():
            if not option.startswith('_'):
                settings[option] = value

        return '<PeekabooConfig(%s)>' % settings

    __repr__ = __str__



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# db.py                                                                       #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

""" A class wrapping database operations needed by Peekaboo based on
SQLAlchemy. """

import threading
import logging
from datetime import datetime, timedelta
from sqlalchemy import Column, Integer, String, Text, DateTime, ForeignKey, \
        Enum, Index
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.engine import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session, relationship
from sqlalchemy.exc import SQLAlchemyError, IntegrityError
from peekaboo import __version__
from peekaboo.ruleset import Result
from peekaboo.exceptions import PeekabooDatabaseError

DB_SCHEMA_VERSION = 7

logger = logging.getLogger(__name__)
Base = declarative_base()


#
# Database schema definition.
##############################################################################


class PeekabooMetadata(Base):
    """ Definition of the _meta table for Peekaboo. """
    __tablename__ = '_meta'
    peekaboo_version = Column(String(10), nullable=False,
                              primary_key=True)
    db_schema_version = Column(Integer, nullable=False,
                               primary_key=True,
                               autoincrement=False)
    cuckoo_version = Column(String(10), nullable=False,
                            primary_key=True)

    def __str__(self):
        return (
            '<PeekabooMetadata(peekaboo_version="%s", db_schema_version="%s", '
            'cuckoo_version="%s")>'
            % (self.peekaboo_version,
               self.db_schema_version,
               self.cuckoo_version)
        )

    __repr__ = __str__


class InFlightSample(Base):
    """
    Table tracking whether a specific sample is currently being analysed and by
    which Peekaboo instance.
    """
    __tablename__ = 'in_flight_samples_v%d' % DB_SCHEMA_VERSION

    # Indices:
    # - general considerations: The table will likely never have more than a
    #   couple of hundret entries. But we add and delete quite frequently.
    # - uniqueness of the primary key ensures atomic insertion when adding a
    #   lock
    # - column: we delete our own stale locks by instance_id.
    # - column: we delete other's stale locks by start_time.
    # - compound: we delete our own locks by sha256sum and instance_id.
    #   (admittedly a bit of overkill since the individual columns are already
    #   indexed.)

    sha256sum = Column(String(64), primary_key=True)
    instance_id = Column(Integer, nullable=False, index=True)
    start_time = Column(DateTime, nullable=False, index=True)

    __table_args__ = (
        # Index names need to be unique per schema in postgresql.
        Index('ix_%s_sha_iid' % __tablename__, sha256sum, instance_id),
        )

    def __str__(self):
        return (
            '<InFlightSample(sha256sum="%s", instance_id="%s", '
            'start_time="%s")>'
            % (self.sha256sum,
               self.instance_id,
               self.start_time.strftime("%Y%m%dT%H%M%S"))
        )

    __repr__ = __str__


class SampleInfo(Base):
    """ Definition of the sample_info table. """
    __tablename__ = 'sample_info_v%d' % DB_SCHEMA_VERSION

    # Indices:
    # - general considerations: The table grows very large over time. Every
    #   sample is checked against it to find a cached analysis result.
    #   Otherwise it's quite unused currently.
    # - compound: we frequently search by sha256sum and file extension

    id = Column(Integer, primary_key=True)
    sha256sum = Column(String(64), nullable=False)
    file_extension = Column(String(16), nullable=True)
    result = Column(Enum(Result), nullable=False)
    reason = Column(Text, nullable=True)

    __table_args__ = (
        Index('ix_%s_sha_fe' % __tablename__, sha256sum, file_extension),
        )

    def __str__(self):
        return ('<SampleInfo(sample_sha256_hash="%s", file_extension="%s", '
                'reason="%s")>'
                % (self.sha256sum,
                   self.file_extension,
                   self.reason))

    __repr__ = __str__


class AnalysisJournal(Base):
    """ Definition of the analysis_jobs table. """
    __tablename__ = 'analysis_jobs_v%d' % DB_SCHEMA_VERSION

    # Indices:
    # - general considerations: We're not using this table at all currently.
    #   All indexes are therefore currenly only futureproofing and admin or
    #   scripting aids.
    # - column: Speed up removal of old entries by analysis_time.
    # - column: Index foreign key to speed up search and removal by sample id.

    id = Column(Integer, primary_key=True)
    job_hash = Column(String(255), nullable=False)
    cuckoo_job_id = Column(Integer, nullable=False)
    filename = Column(String(255), nullable=False)
    analysis_time = Column(DateTime, nullable=False,
                           index=True)
    sample_id = Column(Integer, ForeignKey(SampleInfo.id),
                       nullable=False, index=True)
    sample = relationship('SampleInfo')

    def __str__(self):
        return (
            '<AnalysisJournal(job_hash="%s", cuckoo_job_id="%s", '
            'filename="%s", analysis_time="%s")>'
            % (self.job_hash,
               self.cuckoo_job_id,
               self.filename,
               self.analysis_time.strftime("%Y%m%dT%H%M%S"))
        )

    __repr__ = __str__


#
# End of database schema definition.
##############################################################################


class PeekabooDatabase(object):
    """ Peekaboo's database. """
    def __init__(self, db_url, instance_id=0,
                 stale_in_flight_threshold=15*60,
                 log_level=logging.WARNING):
        """
        Initialize the Peekaboo database handler.

        @param db_url: An RFC 1738 URL that points to the database.
        @param instance_id: A positive, unique ID differentiating this Peekaboo
                            instance from any other instance using the same
                            database for concurrency coordination. Value of 0
                            means that we're alone and have no other instances
                            to worry about.
        @param stale_in_flight_threshold: Number of seconds after which a in
        flight marker is considered stale and deleted or ignored.
        """
        logging.getLogger('sqlalchemy.engine').setLevel(log_level)

        self.__engine = create_engine(db_url, pool_recycle=1)
        session_factory = sessionmaker(bind=self.__engine)
        self.__session = scoped_session(session_factory)
        self.__lock = threading.RLock()
        self.instance_id = instance_id
        self.stale_in_flight_threshold = stale_in_flight_threshold
        if not self._db_schema_exists():
            self._init_db()
            logger.debug('Database schema created.')

    def analysis_save(self, sample):
        """
        Save an analysis task to the analysis journal in the database.

        @param sample: The sample object for this analysis task.
        """
        analysis = AnalysisJournal()
        analysis.job_hash = sample.job_hash
        analysis.cuckoo_job_id = sample.job_id
        analysis.filename = sample.filename
        analysis.analysis_time = datetime.now()
        sample_info = self.sample_info_fetch(sample)
        if sample_info is None:
            sample_info = SampleInfo(
                sha256sum=sample.sha256sum,
                file_extension=sample.file_extension,
                result=sample.result,
                reason=sample.reason)

        analysis.sample = sample_info

        with self.__lock:
            session = self.__session()
            session.add(analysis)
            try:
                session.commit()
            except SQLAlchemyError as error:
                session.rollback()
                raise PeekabooDatabaseError(
                    'Failed to add analysis task to the database: %s' % error
                )
            finally:
                session.close()

    def sample_info_fetch(self, sample):
        """
        Fetch information stored in the database about a given sample object.

        @param sample: The sample object of which the information shall be
                       fetched from the database.
        @return: A SampleInfo object containing the information stored in teh
                 database about the sample.
        """
        with self.__lock:
            session = self.__session()
            sample_info = session.query(SampleInfo).filter_by(
                sha256sum=sample.sha256sum,
                file_extension=sample.file_extension).first()
            session.close()
        return sample_info

    def mark_sample_in_flight(self, sample, instance_id=None, start_time=None):
        """
        Mark a sample as in flight, i.e. being worked on by an instance.

        @param sample: The sample to mark as in flight.
        @param instance_id: (optionally) The ID of the instance that is
                            handling this sample. Default: Us.
        @param start_time: Override the time the marker was placed for
                           debugging purposes.
        """
        # an instance id of 0 denotes that we're alone and don't need to track
        # in-flight samples in the database
        if self.instance_id == 0:
            return True

        # use our own instance id if none is given
        if instance_id is None:
            instance_id = self.instance_id

        if start_time is None:
            start_time = datetime.utcnow()

        session = self.__session()

        # try to mark this sample as in flight in an atomic insert operation
        sha256sum = sample.sha256sum
        session.add(InFlightSample(sha256sum=sha256sum,
                                   instance_id=instance_id,
                                   start_time=start_time))

        locked = False
        try:
            session.commit()
            locked = True
            logger.debug('Marked sample %s as in flight', sha256sum)
        # duplicate primary key == entry already exists
        except IntegrityError:
            session.rollback()
            logger.debug('Sample %s is already in flight on another instance',
                         sha256sum)
        except SQLAlchemyError as error:
            session.rollback()
            raise PeekabooDatabaseError('Unable to mark sample as in flight: '
                                        '%s' % error)
        finally:
            session.close()

        return locked

    def clear_sample_in_flight(self, sample, instance_id=None):
        """
        Clear the mark that a sample is being processed by an instance.

        @param sample: The sample to clear from in-flight list.
        @param instance_id: (optionally) The ID of the instance that is
                            handling this sample. Default: Us.
        """
        # an instance id of 0 denotes that we're alone and don't need to track
        # in-flight samples in the database
        if self.instance_id == 0:
            return

        # use our own instance id if none is given
        if instance_id is None:
            instance_id = self.instance_id

        session = self.__session()

        # clear in-flight marker from database
        sha256sum = sample.sha256sum
        query = session.query(InFlightSample).filter(
            InFlightSample.sha256sum == sha256sum).filter(
                InFlightSample.instance_id == instance_id)

        try:
            # delete() is not queued and goes to the DB before commit()
            cleared = query.delete()
            session.commit()
        except SQLAlchemyError as error:
            session.rollback()
            raise PeekabooDatabaseError('Unable to clear in-flight status of '
                                        'sample: %s' % error)
        finally:
            session.close()

        if cleared == 0:
            raise PeekabooDatabaseError('Unexpected inconsistency: Sample %s '
                                        'not recoreded as in-flight upon '
                                        'clearing flag.' % sha256sum)
        elif cleared > 1:
            raise PeekabooDatabaseError('Unexpected inconsistency: Multiple '
                                        'instances of sample %s in-flight '
                                        'status cleared against database '
                                        'constraints!?' % sha256sum)

        logger.debug('Cleared sample %s from in-flight list', sha256sum)

    def clear_in_flight_samples(self, instance_id=None):
        """
        Clear all in-flight markers left over by previous runs or other
        instances by removing them from the lock table.

        @param instance_id: Clear our own (None), another instance's (positive
                            integer) or all instances' (negative integer) locks.
                            Since an instance_id of 0 disables in-flight sample
                            tracking, no instance will ever set a marker with
                            that ID so that specifying 0 here will amount to a
                            no-op or rather clean-up of invalid entries.
        """
        # an instance id of 0 denotes that we're alone and don't need to track
        # in-flight samples
        if self.instance_id == 0:
            return

        # use our own instance id if none is given
        if instance_id is None:
            instance_id = self.instance_id

        session = self.__session()

        if instance_id < 0:
            # delete all locks
            query = session.query(InFlightSample)
            logger.debug('Clearing database of all in-flight samples.')
        else:
            # delete only the locks of a specific instance
            query = session.query(InFlightSample).filter(
                InFlightSample.instance_id == instance_id)
            logger.debug('Clearing database of all in-flight samples of '
                         'instance %d.', instance_id)
        try:
            # delete() is not queued and goes to the DB before commit()
            query.delete()
            session.commit()
        except SQLAlchemyError as error:
            session.rollback()
            raise PeekabooDatabaseError('Unable to clear the database of '
                                        'in-flight samples: %s' % error)
        finally:
            session.close()

    def clear_stale_in_flight_samples(self):
        """
        Clear all in-flight markers that are too old and therefore stale. This
        detects instances which are locked up, crashed or shut down.
        """
        # an instance id of 0 denotes that we're alone and don't need to track
        # in-flight samples in the database
        if self.instance_id == 0:
            return True

        session = self.__session()

        # delete only the locks of a specific instance
        query = session.query(InFlightSample).filter(
            InFlightSample.start_time <= datetime.utcnow() - timedelta(
                seconds=self.stale_in_flight_threshold))
        logger.debug(
            'Clearing database of all stale in-flight samples '
            '(%d seconds)', self.stale_in_flight_threshold)

        try:
            # the loop triggers the query, so only do it if debugging is
            # enabled
            if logger.isEnabledFor(logging.DEBUG):
                # obviously there's a race between logging and actual delete
                # here, use with caution, compare with actual number of markers
                # cleared below before relying on it for debugging
                for stale in query:
                    logger.debug('Stale in-flight marker to clear: %s', stale)

            # delete() is not queued and goes to the DB before commit()
            cleared = query.delete()
            session.commit()
            if cleared > 0:
                logger.warn('%d stale in-flight samples cleared.', cleared)
        except SQLAlchemyError as error:
            session.rollback()
            raise PeekabooDatabaseError('Unable to clear the database of '
                                        'stale in-flight samples: %s' % error)
        finally:
            session.close()

        return cleared > 0

    def drop(self):
        """ Drop all tables of the database. """
        try:
            Base.metadata.drop_all(self.__engine)
        except SQLAlchemyError as error:
            raise PeekabooDatabaseError(
                'Unable to drop all tables of the database: %s' % error)

    def _db_schema_exists(self):
        if not self.__engine.dialect.has_table(self.__engine, '_meta'):
            return False

        session = self.__session()
        # get the first of potentially multiple metadata entries ordered by
        # descending schema version to get the newest currently configured
        # schema
        meta = session.query(PeekabooMetadata).order_by(
            PeekabooMetadata.db_schema_version.desc()).first()
        if not meta:
            return False

        schema_version = meta.db_schema_version
        session.close()
        if schema_version < DB_SCHEMA_VERSION:
            logger.info('Adding new database schema.')
            return False

        return True

    def _init_db(self):
        """
        Initializes the Peekaboo database by creating tables and
        writing meta information to the '_meta' table.
        """
        Base.metadata.create_all(self.__engine)
        meta = PeekabooMetadata()
        meta.peekaboo_version = __version__
        meta.db_schema_version = DB_SCHEMA_VERSION
        # TODO: Get Cuckoo version.
        meta.cuckoo_version = '2.0'
        session = self.__session()
        session.add(meta)
        try:
            session.commit()
        except SQLAlchemyError as error:
            session.rollback()
            raise PeekabooDatabaseError(
                'Cannot initialize the database: %s' % error)
        finally:
            session.close()



#!/usr/bin/env python

###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# setup.py                                                                    #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################


from codecs import open
from os import path
from sys import path as pythonpath
from setuptools import setup, find_packages

# Add peekaboo to PYTHONPATH
pythonpath.append(path.dirname(path.dirname(path.abspath(__file__))))
from peekaboo import __version__, __author__, __license__, __description__

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# get the dependencies and installs
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [x.strip() for x in all_reqs if 'git+' not in x]
dependency_links = [x.strip().replace('git+', '')
                    for x in all_reqs if 'git+' not in x]

setup(
    name='PeekabooAV',
    version=__version__,
    description=__description__,
    long_description=long_description,
    url='https://github.com/scVENUS/PeekabooAV.git',
    download_url='https://github.com/scVENUS/PeekabooAV/archive/master.zip',
    license=__license__,
    classifiers=[
      'Development Status :: 4 - Beta',
      'Operating System :: POSIX',
      'Programming Language :: Python',
      'Programming Language :: Python :: 2',
      'Programming Language :: Python :: 3',
      'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
      'Natural Language :: English',
      'Natural Language :: German',
      'Topic :: Communications :: Email :: Filters',
    ],
    keywords='Cuckoo, Amavis',
    packages=find_packages(exclude=['docs', 'tests*']),
    include_package_data=True,
    # package_files augments MANIFEST.in in what is packaged into a
    # distribution. Files to add must be inside a package. Thus files in the
    # root of our source directory cannot be packaged with this. Files inside
    # packages will stay there, totally obscured to the user. Meant for default
    # configuration or other package-internal data.
    #  package_files=[...],
    #
    # data_files is another way to augment what is installed from the
    # distribution. Allows paths outside packages for both sources and targets.
    # Absolute paths are strongly discouraged because they exhibit confusing if
    # not downright broken behaviour. Relative paths are added to the
    # distribution and then installed into
    # site-packages/<package>-<ver>.egg/<relative path> (setup.py) or <python
    # prefix>/<relative path> (pip). The latter works well with venvs and
    # distribution-provided pip (e.g. /usr/local/<relative path>).
    # We go this route for providing sample files because using setup.py
    # directly is discouraged anyway and "only" tucks the files away in the egg
    # directory, providing the most consistent option.
    data_files=[
        ("share/doc/peekaboo", [
            "README.md",
            "CHANGELOG.md",
            "peekaboo.conf.sample",
            "ruleset.conf.sample",
        ]),
        ("share/doc/peekaboo/systemd", [
            "systemd/peekaboo.service",
        ]),
        ("share/doc/peekaboo/amavis", [
            "amavis/10-ask_peekaboo",
        ]),
    ],
    # overriding the whole install_data command allows for arbitrary
    # installation mechanics but does not solve the problem of adding files to
    # a binary distribution (e.g. wheel, which pip uses internally always) in
    # such a way that they will later be put at the correct location. Thus they
    # would go around pip entirely, be missing from any actual wheel
    # distribution package, pollute the system directly and not be removed upon
    # uninstall or upgrade.
    #  cmdclass={
    #      'install_data': OffsetDataInstall,
    #  },
    author=__author__,
    install_requires=install_requires,
    dependency_links=dependency_links,
    author_email='felix.bauer@atos.net',
    entry_points={
        'console_scripts': [
            'peekaboo = peekaboo.daemon:run',
            'peekaboo-util = peekaboo.cli:main',
        ],
    }
)



###############################################################################
#                                                                             #
# Peekaboo Extended Email Attachment Behavior Observation Owl                 #
#                                                                             #
# ruleset/                                                                    #
#         rules.py                                                            #
###############################################################################
#                                                                             #
# Copyright (C) 2016-2020  science + computing ag                             #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or (at       #
# your option) any later version.                                             #
#                                                                             #
# This program is distributed in the hope that it will be useful, but         #
# WITHOUT ANY WARRANTY; without even the implied warranty of                  #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           #
# General Public License for more details.                                    #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

""" Classes implementing the Ruleset """


import re
import logging
from peekaboo.ruleset import Result, RuleResult
from peekaboo.ruleset.expressions import ExpressionParser, \
        IdentifierMissingException
from peekaboo.exceptions import PeekabooAnalysisDeferred, \
        CuckooSubmitFailedException, PeekabooRulesetConfigError
from peekaboo.sample import Sample
from peekaboo.toolbox.cuckoo import CuckooReport
from peekaboo.toolbox.ole import Oletools, OletoolsReport
from peekaboo.toolbox.file import Filetools, FiletoolsReport


logger = logging.getLogger(__name__)


class Rule(object):
    """ This is the base class for all rules. It provides common infrastructure
    such as resources that can be used by the rules (configuration, database
    connection) or helper functions. """
    rule_name = 'unimplemented'

    def __init__(self, config, db_con):
        """ Initialize common configuration and resources """
        self.db_con = db_con
        self.config = config

        # initialise and validate configuration
        self.config_options = {}
        self.get_config()
        # if this rule has (tried to) read any options from the config, it must
        # believe them to be known and allowed
        if self.config_options:
            self.config.check_section_options(
                self.rule_name, self.config_options.keys())

    def result(self, result, reason, further_analysis):
        """ Construct a RuleResult for returning to the engine. """
        return RuleResult(self.rule_name, result=result, reason=reason,
                          further_analysis=further_analysis)

    def evaluate(self, sample):
        """ Evaluate a rule agaimst a sample. Findings are recorded in the
        sample or returned in the rule result. *Must not* change the rule
        object's internal state because it will be called by multiple workers
        in parallel.

        @param sample: The sample to evaluate.
        @returns: RuleResult containing verdict, reason, source of this
                  assessment (i.e. the rule's name) and whether to continue
                  analysis or not.
        """
        raise NotImplementedError

    def get_config(self):
        """ Extract this rule's configuration out of the ruleset configuration
        object given at creation. To be overridden by child classes if they
        have configuration options. """
        # pass

    def get_config_value(self, option, default, option_type=None):
        """ Get a configuation value for this rule from the ruleset
        configuration. Getter routine and option name to be provided by caller.
        The rule's name is always used as configuration section name.

        @param option: name of option to read
        @type option: string
        @param default: default value to use as fallback and for type
                        determination
        @type default: None, int, float, string, list, tuple
        @param option_type: force the option's value type, necessary for lists
                            of regular expressions or log levels by specifying
                            self.config.RELIST or self.config.LOG_LEVEL
        @type option_type: option type constant of PeekabooConfigParser, e.g.
                           LOG_LEVEL
        @param args, kwargs: additional arguments passed to the getter routine,
                             such as fallback.

        @returns: configuration value read from config
        """
        # mark this config option as known
        self.config_options[option] = True
        return self.config.get_by_type(
            self.rule_name, option, fallback=default, option_type=option_type)

    def get_cuckoo_report(self, sample):
        """ Get the samples cuckoo_report or submit the sample for analysis by
            Cuckoo.

            @returns: CuckooReport
        """
        if sample.cuckoo_failed:
            return None

        report = sample.cuckoo_report
        if report is not None:
            return report

        try:
            job_id = sample.submit_to_cuckoo()
        except CuckooSubmitFailedException as failed:
            logger.error("Submit to Cuckoo failed: %s", failed)
            return None

        logger.info('Sample submitted to Cuckoo. Job ID: %s. '
                    'Sample: %s', job_id, sample)
        raise PeekabooAnalysisDeferred()

    def get_oletools_report(self, sample):
        """ Get an Oletools report on the sample.

        @returns: OletoolsReport
        """
        return Oletools(sample).get_report()

    def get_filetools_report(self, sample):
        """ Get a Filetools report on the sample.

        @returns: FiletoolsReport
        """
        return Filetools(sample).get_report()


class KnownRule(Rule):
    """ A rule determining if a sample is known by looking at the database for
    a previous record of an identical sample sample. """
    rule_name = 'known'

    def evaluate(self, sample):
        """ Try to get information about the sample from the database. Return
        the old result and reason if found and advise the engine to stop
        processing. """
        sample_info = self.db_con.sample_info_fetch(sample)
        if sample_info:
            return self.result(sample_info.result, sample_info.reason, False)

        return self.result(Result.unknown,
                           _("File is not yet known to the system"),
                           True)


class FileLargerThanRule(Rule):
    """ A rule determining by file size whether a sample can be harmful at all.
    """
    rule_name = 'file_larger_than'

    def get_config(self):
        self.size_threshold = self.get_config_value('bytes', 5)

    def evaluate(self, sample):
        """ Evaluate whether the sample is larger than a certain threshold.
        Advise the engine to stop processing if the size is below the
        threshold. """
        try:
            sample_size = sample.file_size
        except OSError as oserr:
            return self.result(
                Result.failed,
                _("Failure to determine sample file size: %s") % oserr,
                False)

        if sample_size > self.size_threshold:
            return self.result(Result.unknown,
                               _("File has more than %d bytes")
                               % self.size_threshold,
                               True)

        return self.result(
            Result.ignored,
            _("File is only %d bytes long") % sample_size,
            False)


class FileTypeOnWhitelistRule(Rule):
    """ A rule checking whether the known file type(s) of the sample are on a
    whitelist. """
    rule_name = 'file_type_on_whitelist'

    def get_config(self):
        whitelist = self.get_config_value('whitelist', [])
        if not whitelist:
            raise PeekabooRulesetConfigError(
                "Empty whitelist, check %s rule config." % self.rule_name)

        self.whitelist = set(whitelist)

    def evaluate(self, sample):
        """ Ignore the file only if *all* of its mime types are on the
        whitelist and we could determine at least one. """
        filereport = self.get_filetools_report(sample)
        mimetypes = filereport.mime_types
        if sample.type_declared is not None:
            mimetypes.add(sample.type_declared)
        if mimetypes and mimetypes.issubset(self.whitelist):
            return self.result(Result.ignored,
                               _("File type is on whitelist"),
                               False)

        return self.result(Result.unknown,
                           _("File type is not on whitelist"),
                           True)


class FileTypeOnGreylistRule(Rule):
    """ A rule checking whether any of the sample's known file types are on a
    greylist, i.e. enabled for analysis. """
    rule_name = 'file_type_on_greylist'

    def get_config(self):
        greylist = self.get_config_value('greylist', [])
        if not greylist:
            raise PeekabooRulesetConfigError(
                "Empty greylist, check %s rule config." % self.rule_name)

        self.greylist = set(greylist)

    def evaluate(self, sample):
        """ Continue analysis if any of the sample's MIME types are on the
        greylist or in case we don't have one. """
        filereport = self.get_filetools_report(sample)
        mimetypes = filereport.mime_types
        if sample.type_declared is not None:
            mimetypes.add(sample.type_declared)
        if not mimetypes or mimetypes.intersection(self.greylist):
            return self.result(Result.unknown,
                               _("File type is on the list of types to "
                                 "analyze"),
                               True)

        return self.result(Result.unknown,
                           _("File type is not on the list of types to "
                             "analyse (%s)") % mimetypes,
                           False)


class OleRule(Rule):
    """ A common base class for rules that evaluate the Ole report. """
    def evaluate(self, sample):
        """ Report the sample as bad if it contains a macro. """
        # we always get a report, albeit a maybe empty one
        return self.evaluate_report(self.get_oletools_report(sample))

    def evaluate_report(self, report):
        """ Evaluate an Ole report.

        @param report: The Ole report.
        @returns: RuleResult containing verdict.
        """
        raise NotImplementedError


class OfficeMacroRule(OleRule):
    """ A rule checking the sample for Office macros. """
    rule_name = 'office_macro'

    def evaluate_report(self, report):
        """ Report the sample as bad if it contains a macro. """
        if report.has_office_macros:
            return self.result(Result.bad,
                               _("The file contains an Office macro"),
                               False)

        return self.result(Result.unknown,
                           _("The file does not contain a recognizable "
                             "Office macro"),
                           True)


class OfficeMacroWithSuspiciousKeyword(OleRule):
    """ A rule checking the sample for Office macros. """
    rule_name = 'office_macro_with_suspicious_keyword'

    def get_config(self):
        # get list of keywords from config file
        self.suspicious_keyword_list = self.get_config_value(
            'keyword', [], option_type=self.config.IRELIST)
        if not self.suspicious_keyword_list:
            raise PeekabooRulesetConfigError(
                "Empty suspicious keyword list, check %s rule config." %
                self.rule_name)

    def evaluate_report(self, report):
        if report.has_office_macros_with_suspicious_keyword(self.suspicious_keyword_list):
            return self.result(Result.bad,
                               _("The file contains an Office macro which "
                                 "runs at document open"),
                               False)

        return self.result(Result.unknown,
                           _("The file does not contain a recognizable "
                             "Office macro that is run at document open"),
                           True)


class CuckooRule(Rule):
    """ A common base class for rules that evaluate the Cuckoo report. """
    def evaluate(self, sample):
        """ If a report is present for the sample in question we call method
        evaluate_report() implemented by subclasses to evaluate it for
        findings. Otherwise we submit the sample to Cuckoo and raise
        PeekabooAnalysisDeferred to abort the current run of the ruleset
        until the report arrives. If submission to Cuckoo fails we will
        ourselves report the sample as failed.

        @param sample: The sample to evaluate.
        @raises PeekabooAnalysisDeferred: if the sample was submitted to Cuckoo
        @returns: RuleResult containing verdict.
        """
        report = self.get_cuckoo_report(sample)
        if report is None:
            # exception message intentionally not present in message
            # delivered back to client as to not disclose internal
            # information, should request user to contact admin instead
            return self.result(
                Result.failed,
                _("Behavioral analysis by Cuckoo has produced an error "
                  "and did not finish successfully"),
                False)

        # call report evaluation function if we get here
        return self.evaluate_report(report)

    def evaluate_report(self, report):
        """ Evaluate a Cuckoo report.

        @param report: The Cuckoo report.
        @returns: RuleResult containing verdict.
        """
        raise NotImplementedError


class CuckooEvilSigRule(CuckooRule):
    """ A rule evaluating the signatures from the Cuckoo report against a list
    of signatures considered bad. """
    rule_name = 'cuckoo_evil_sig'

    def get_config(self):
        # list all installed signatures
        # grep -o "description.*" -R . ~/cuckoo2.0/modules/signatures/
        self.bad_sigs = self.get_config_value(
            'signature', [], option_type=self.config.RELIST)
        if not self.bad_sigs:
            raise PeekabooRulesetConfigError(
                "Empty bad signature list, check %s rule config." %
                self.rule_name)

    def evaluate_report(self, report):
        """ Evaluate the sample against signatures that if matched mark a
        sample as bad. """
        # look through matched signatures
        sigs = []
        for descr in report.signatures:
            logger.debug("Signature from cuckoo report: %s",
                         descr['description'])
            sigs.append(descr['description'])

        # check if there is a "bad" signatures and return bad
        matched_bad_sigs = []
        for bad_sig in self.bad_sigs:
            # iterate over each sig individually to allow regexes to use
            # anchors such as ^ and $ and avoid mismatches, e.g. by ['foo',
            # 'bar'] being stringified to "['foo', 'bar']" and matching
            # /fo.*ar/.
            for sig in sigs:
                match = re.search(bad_sig, sig)
                if match:
                    matched_bad_sigs.append(sig)

        if not matched_bad_sigs:
            return self.result(Result.unknown,
                               _("No signature suggesting malware detected"),
                               True)

        matched = ''.ljust(8).join(["%s\n" % s for s in matched_bad_sigs])
        return self.result(Result.bad,
                           _("The following signatures have been recognized: "
                             "%s") % matched,
                           False)


class CuckooScoreRule(CuckooRule):
    """ A rule checking the score reported by Cuckoo against a configurable
    threshold. """
    rule_name = 'cuckoo_score'

    def get_config(self):
        self.score_threshold = self.get_config_value('higher_than', 4.0)

    def evaluate_report(self, report):
        """ Evaluate the score reported by Cuckoo against the threshold from
        the configuration and report sample as bad if above. """

        if report.score >= self.score_threshold:
            return self.result(Result.bad,
                               _("Cuckoo score >= %s: %s") %
                               (self.score_threshold, report.score),
                               False)

        return self.result(Result.unknown,
                           _("Cuckoo score < %s: %s") %
                           (self.score_threshold, report.score),
                           True)


class RequestsEvilDomainRule(CuckooRule):
    """ A rule checking the domains reported as requested by the sample by
    Cuckoo against a blacklist. """
    rule_name = 'requests_evil_domain'

    def get_config(self):
        self.evil_domains = self.get_config_value('domain', [])
        if not self.evil_domains:
            raise PeekabooRulesetConfigError(
                "Empty evil domain list, check %s rule config."
                % self.rule_name)

    def evaluate_report(self, report):
        """ Report the sample as bad if one of the requested domains is on our
        list of evil domains. """

        for domain in report.requested_domains:
            if domain in self.evil_domains:
                return self.result(Result.bad,
                                   _("The file attempts to contact at least "
                                     "one domain on the blacklist (%s)")
                                   % domain,
                                   False)

        return self.result(Result.unknown,
                           _("File does not seem to attempt contact with "
                             "domains on the blacklist"),
                           True)


class CuckooAnalysisFailedRule(CuckooRule):
    """ A rule checking the final status reported by Cuckoo for success. """
    rule_name = 'cuckoo_analysis_failed'

    def get_config(self):
        self.failure_matches = self.get_config_value('failure', [])
        self.success_matches = self.get_config_value(
            'success', ['analysis completed successfully'])

    def evaluate_report(self, report):
        """ Report the sample as bad if the Cuckoo indicates that the analysis
        has failed. """
        if report.errors:
            logger.warning('Cuckoo produced %d error(s) during processing.',
                           len(report.errors))

        failure_reason = _("Behavioral analysis by Cuckoo has produced "
                           "an error and did not finish successfully")

        for entry in report.cuckoo_server_messages:
            for failure in self.failure_matches:
                if failure in entry:
                    logger.debug('Failure indicator "%s" found in Cuckoo '
                                 'messages', failure)
                    return self.result(Result.failed, failure_reason, False)

        for entry in report.cuckoo_server_messages:
            for success in self.success_matches:
                if success in entry:
                    logger.debug('Success indicator "%s" found in Cuckoo '
                                 'messages', success)
                    return self.result(Result.unknown,
                                       _("Behavioral analysis by Cuckoo "
                                         "completed successfully"),
                                       True)

        logger.debug('Neither success nor failure indicators found, '
                     'considering analysis failed.')
        return self.result(Result.failed, failure_reason, False)


class ExpressionRule(Rule):
    """ A rule checking the sample and cuckoo report against an almost
    arbitrary logical expression. """
    rule_name = 'expressions'

    def get_config(self):
        # allow to raise debug level of expressions module explicitly. Beware:
        # This affects not just individual objects but the whole module which
        # is why we do it by poking the logger and not via a setter method.
        log_level = self.get_config_value(
            'log_level', logging.WARNING, option_type=self.config.LOG_LEVEL)
        expression_logger = logging.getLogger('peekaboo.ruleset.expressions')
        expression_logger.setLevel(log_level)

        self.expressions = self.get_config_value('expression', [])
        if not self.expressions:
            raise PeekabooRulesetConfigError(
                "List of expressions empty, check %s rule config."
                % self.rule_name)

        self.rules = []
        parser = ExpressionParser()

        # context of dummy objects to test rules against
        context = {
            'variables': {
                'sample': Sample("dummy"),
                'cuckooreport': CuckooReport(),
                'olereport': OletoolsReport(),
                'filereport': FiletoolsReport(),
            }
        }

        for expr in self.expressions:
            try:
                rule = parser.parse(expr)
                logger.debug("Expression from config file: %s", expr)
                logger.debug("Expression parsed: %s", rule)
            except SyntaxError as error:
                raise PeekabooRulesetConfigError(error)

            if not rule.is_implication():
                raise PeekabooRulesetConfigError(
                    "Malformed expression, missing implication: %s" % expr)

            # run expression against dummy objects to find out if it's
            # attempting anything illegal
            try:
                rule.eval(context=context)
            except IdentifierMissingException as missing:
                # our dummy context provides everything we would provide at
                # runtime as well, so any missing identifier is an error at
                # this point
                identifier = missing.name
                raise PeekabooRulesetConfigError(
                    "Invalid expression, unknown identifier %s: %s" % (
                        identifier, expr))
            except AttributeError as missing:
                raise PeekabooRulesetConfigError(
                    "Invalid expression, %s: %s" % (missing, expr))

            self.rules.append(rule)

    def evaluate(self, sample):
        """ Match what rules report against our known result status names. """
        for i, rule in enumerate(self.rules):
            result = None
            context = {'variables': {'sample': sample}}

            # retry until rule evaluation doesn't throw exceptions any more
            while True:
                try:
                    result = rule.eval(context=context)
                    break
                except IdentifierMissingException as missing:
                    identifier = missing.name

                if identifier == "cuckooreport":
                    logger.debug("Expression requests cuckoo report")
                    value = self.get_cuckoo_report(sample)
                    if value is None:
                        return self.result(
                            Result.failed,
                            _("Evaluation of expression couldn't get cuckoo "
                              "report."),
                            False)
                elif identifier == "olereport":
                    logger.debug("Expression requests oletools report")
                    value = self.get_oletools_report(sample)
                elif identifier == "filereport":
                    logger.debug("Expression requests filetools report")
                    value = self.get_filetools_report(sample)
                # elif here for other identifiers
                else:
                    return self.result(
                        Result.failed,
                        _("Evaluation of expression uses undefined "
                          "identifier."),
                        False)

                context['variables'][identifier] = value
                # beware: here we intentionally loop on through for retry

            # our implication returns None if expression did not match
            if result is None:
                continue

            # eval will return something completely different if implication is
            # missing
            if not isinstance(result, Result):
                logger.warning("Expression %d is returning an invalid result, "
                               "failing evaluation: %s", i, rule)
                result = Result.failed

            return self.result(
                result,
                _("The expression (%d) classified the sample as %s")
                % (i, result),
                False)

        return self.result(
            Result.unknown,
            _("No expression classified the sample in any way."),
            True)


class FinalRule(Rule):
    """ A catch-all rule. """
    rule_name = 'final_rule'

    def evaluate(self, sample):
        """ Report an unknown analysis result indicating that nothing much can
        be said about the sample. """
        return self.result(Result.unknown,
                           _("File does not seem to exhibit recognizable "
                             "malicious behaviour"),
                           True)
